from collections import defaultdict
from dataclasses import asdict, dataclass
from datetime import datetime
from typing import TYPE_CHECKING, Annotated, Optional

import pandas as pd
import strawberry
from openinference.semconv.trace import SpanAttributes
from sqlalchemy import select
from strawberry import UNSET, Info, lazy
from strawberry.relay import Connection, Node, NodeID

from phoenix.db import models
from phoenix.server.api.context import Context
from phoenix.server.api.input_types.AnnotationFilter import AnnotationFilter, satisfies_filter
from phoenix.server.api.types.AnnotationSummary import AnnotationSummary
from phoenix.server.api.types.CostBreakdown import CostBreakdown
from phoenix.server.api.types.MimeType import MimeType
from phoenix.server.api.types.pagination import ConnectionArgs, CursorString, connection_from_list
from phoenix.server.api.types.SpanCostDetailSummaryEntry import SpanCostDetailSummaryEntry
from phoenix.server.api.types.SpanCostSummary import SpanCostSummary
from phoenix.server.api.types.SpanIOValue import SpanIOValue
from phoenix.server.api.types.TokenUsage import TokenUsage

if TYPE_CHECKING:
    from phoenix.server.api.types.Project import Project
    from phoenix.server.api.types.ProjectSessionAnnotation import ProjectSessionAnnotation
    from phoenix.server.api.types.Trace import Trace


@strawberry.type
class ProjectSession(Node):
    id: NodeID[int]
    db_record: strawberry.Private[Optional[models.ProjectSession]] = None

    def __post_init__(self) -> None:
        if self.db_record and self.id != self.db_record.id:
            raise ValueError("ProjectSession ID mismatch")

    @strawberry.field
    async def session_id(
        self,
        info: Info[Context, None],
    ) -> str:
        if self.db_record:
            val = self.db_record.session_id
        else:
            val = await info.context.data_loaders.project_session_fields.load(
                (self.id, models.ProjectSession.session_id),
            )
        return val

    @strawberry.field
    async def start_time(
        self,
        info: Info[Context, None],
    ) -> datetime:
        if self.db_record:
            val = self.db_record.start_time
        else:
            val = await info.context.data_loaders.project_session_fields.load(
                (self.id, models.ProjectSession.start_time),
            )
        return val

    @strawberry.field
    async def end_time(
        self,
        info: Info[Context, None],
    ) -> datetime:
        if self.db_record:
            val = self.db_record.end_time
        else:
            val = await info.context.data_loaders.project_session_fields.load(
                (self.id, models.ProjectSession.end_time),
            )
        return val

    @strawberry.field
    async def project(
        self,
        info: Info[Context, None],
    ) -> Annotated["Project", lazy(".Project")]:
        from phoenix.server.api.types.Project import Project

        if self.db_record:
            project_rowid = self.db_record.project_id
        else:
            project_rowid = await info.context.data_loaders.project_session_fields.load(
                (self.id, models.ProjectSession.project_id),
            )
        return Project(id=project_rowid)

    @strawberry.field
    async def num_traces(
        self,
        info: Info[Context, None],
    ) -> int:
        return await info.context.data_loaders.session_num_traces.load(self.id)

    @strawberry.field
    async def num_traces_with_error(
        self,
        info: Info[Context, None],
    ) -> int:
        return await info.context.data_loaders.session_num_traces_with_error.load(self.id)

    @strawberry.field
    async def first_input(
        self,
        info: Info[Context, None],
    ) -> Optional[SpanIOValue]:
        record = await info.context.data_loaders.session_first_inputs.load(self.id)
        if record is None:
            return None
        return SpanIOValue(
            mime_type=MimeType(record.mime_type.value),
            cached_value=record.value,
        )

    @strawberry.field
    async def last_output(
        self,
        info: Info[Context, None],
    ) -> Optional[SpanIOValue]:
        record = await info.context.data_loaders.session_last_outputs.load(self.id)
        if record is None:
            return None
        return SpanIOValue(
            mime_type=MimeType(record.mime_type.value),
            cached_value=record.value,
        )

    @strawberry.field
    async def token_usage(
        self,
        info: Info[Context, None],
    ) -> TokenUsage:
        usage = await info.context.data_loaders.session_token_usages.load(self.id)
        return TokenUsage(
            prompt=usage.prompt,
            completion=usage.completion,
        )

    @strawberry.field
    async def traces(
        self,
        info: Info[Context, None],
        first: Optional[int] = 50,
        last: Optional[int] = UNSET,
        after: Optional[CursorString] = UNSET,
        before: Optional[CursorString] = UNSET,
    ) -> Connection[Annotated["Trace", lazy(".Trace")]]:
        from phoenix.server.api.types.Trace import Trace

        args = ConnectionArgs(
            first=first,
            after=after if isinstance(after, CursorString) else None,
            last=last,
            before=before if isinstance(before, CursorString) else None,
        )
        stmt = (
            select(models.Trace)
            .filter_by(project_session_rowid=self.id)
            .order_by(models.Trace.start_time)
        )
        async with info.context.db() as session:
            traces = await session.stream_scalars(stmt)
            data = [Trace(id=trace.id, db_record=trace) async for trace in traces]
        return connection_from_list(data=data, args=args)

    @strawberry.field
    async def trace_latency_ms_quantile(
        self,
        info: Info[Context, None],
        probability: float,
    ) -> Optional[float]:
        return await info.context.data_loaders.session_trace_latency_ms_quantile.load(
            (self.id, probability)
        )

    @strawberry.field
    async def cost_summary(
        self,
        info: Info[Context, None],
    ) -> SpanCostSummary:
        loader = info.context.data_loaders.span_cost_summary_by_project_session
        summary = await loader.load(self.id)
        return SpanCostSummary(
            prompt=CostBreakdown(
                tokens=summary.prompt.tokens,
                cost=summary.prompt.cost,
            ),
            completion=CostBreakdown(
                tokens=summary.completion.tokens,
                cost=summary.completion.cost,
            ),
            total=CostBreakdown(
                tokens=summary.total.tokens,
                cost=summary.total.cost,
            ),
        )

    @strawberry.field
    async def cost_detail_summary_entries(
        self,
        info: Info[Context, None],
    ) -> list[SpanCostDetailSummaryEntry]:
        loader = info.context.data_loaders.span_cost_detail_summary_entries_by_project_session
        summary = await loader.load(self.id)
        return [
            SpanCostDetailSummaryEntry(
                token_type=entry.token_type,
                is_prompt=entry.is_prompt,
                value=CostBreakdown(
                    tokens=entry.value.tokens,
                    cost=entry.value.cost,
                ),
            )
            for entry in summary
        ]

    @strawberry.field
    async def session_annotations(
        self,
        info: Info[Context, None],
    ) -> list[Annotated["ProjectSessionAnnotation", lazy(".ProjectSessionAnnotation")]]:
        """Get all annotations for this session."""
        from .ProjectSessionAnnotation import ProjectSessionAnnotation

        stmt = select(models.ProjectSessionAnnotation).filter_by(project_session_id=self.id)
        async with info.context.db() as session:
            annotations = await session.stream_scalars(stmt)
            return [
                ProjectSessionAnnotation(id=annotation.id, db_record=annotation)
                async for annotation in annotations
            ]

    @strawberry.field(
        description="Summarizes each annotation (by name) associated with the session"
    )  # type: ignore
    async def session_annotation_summaries(
        self,
        info: Info[Context, None],
        filter: Optional[AnnotationFilter] = None,
    ) -> list[AnnotationSummary]:
        """
        Retrieves and summarizes annotations associated with this span.

        This method aggregates annotation data by name and label, calculating metrics
        such as count of occurrences and sum of scores. The results are organized
        into a structured format that can be easily converted to a DataFrame.

        Args:
            info: GraphQL context information
            filter: Optional filter to apply to annotations before processing

        Returns:
            A list of AnnotationSummary objects, each containing:
            - name: The name of the annotation
            - data: A list of dictionaries with label statistics
        """
        # Load all annotations for this span from the data loader
        annotations = await info.context.data_loaders.session_annotations_by_session.load(self.id)

        # Apply filter if provided to narrow down the annotations
        if filter:
            annotations = [
                annotation for annotation in annotations if satisfies_filter(annotation, filter)
            ]

        @dataclass
        class Metrics:
            record_count: int = 0
            label_count: int = 0
            score_sum: float = 0
            score_count: int = 0

        summaries: defaultdict[str, defaultdict[Optional[str], Metrics]] = defaultdict(
            lambda: defaultdict(Metrics)
        )
        for annotation in annotations:
            metrics = summaries[annotation.name][annotation.label]
            metrics.record_count += 1
            metrics.label_count += int(annotation.label is not None)
            metrics.score_sum += annotation.score or 0
            metrics.score_count += int(annotation.score is not None)

        result: list[AnnotationSummary] = []
        for name, label_metrics in summaries.items():
            rows = [{"label": label, **asdict(metrics)} for label, metrics in label_metrics.items()]
            result.append(AnnotationSummary(name=name, df=pd.DataFrame(rows), simple_avg=True))
        return result


INPUT_VALUE = SpanAttributes.INPUT_VALUE.split(".")
INPUT_MIME_TYPE = SpanAttributes.INPUT_MIME_TYPE.split(".")
OUTPUT_VALUE = SpanAttributes.OUTPUT_VALUE.split(".")
OUTPUT_MIME_TYPE = SpanAttributes.OUTPUT_MIME_TYPE.split(".")



from typing import Optional

import strawberry
from strawberry import UNSET
from strawberry.relay import GlobalID
from strawberry.scalars import JSON


@strawberry.input
class DatasetExampleInput:
    input: JSON
    output: JSON
    metadata: JSON
    span_id: Optional[GlobalID] = UNSET



from __future__ import annotations

import asyncio
import importlib.util
import inspect
import json
import time
from abc import ABC, abstractmethod
from collections.abc import AsyncIterator, Callable, Iterator
from dataclasses import dataclass
from functools import wraps
from typing import TYPE_CHECKING, Any, Hashable, Mapping, MutableMapping, Optional, Union

import wrapt
from openinference.instrumentation import safe_json_dumps
from openinference.semconv.trace import (
    OpenInferenceLLMProviderValues,
    OpenInferenceLLMSystemValues,
    SpanAttributes,
)
from strawberry import UNSET
from strawberry.scalars import JSON as JSONScalarType
from typing_extensions import TypeAlias, assert_never, override

from phoenix.config import getenv
from phoenix.evals.models.rate_limiters import (
    AsyncCallable,
    GenericType,
    ParameterSpec,
    RateLimiter,
    RateLimitError,
)
from phoenix.server.api.exceptions import BadRequest
from phoenix.server.api.helpers.playground_registry import PROVIDER_DEFAULT, register_llm_client
from phoenix.server.api.input_types.GenerativeModelInput import GenerativeModelInput
from phoenix.server.api.input_types.InvocationParameters import (
    BoundedFloatInvocationParameter,
    CanonicalParameterName,
    FloatInvocationParameter,
    IntInvocationParameter,
    InvocationParameter,
    InvocationParameterInput,
    JSONInvocationParameter,
    StringInvocationParameter,
    StringListInvocationParameter,
    extract_parameter,
    validate_invocation_parameters,
)
from phoenix.server.api.types.ChatCompletionMessageRole import ChatCompletionMessageRole
from phoenix.server.api.types.ChatCompletionSubscriptionPayload import (
    FunctionCallChunk,
    TextChunk,
    ToolCallChunk,
)
from phoenix.server.api.types.GenerativeProvider import GenerativeProviderKey

if TYPE_CHECKING:
    import httpx
    from anthropic.types import MessageParam, TextBlockParam, ToolResultBlockParam
    from botocore.awsrequest import AWSPreparedRequest  # type: ignore[import-untyped]
    from google.generativeai.types import ContentType
    from openai import AsyncAzureOpenAI, AsyncOpenAI
    from openai.types import CompletionUsage
    from openai.types.chat import ChatCompletionMessageParam, ChatCompletionMessageToolCallParam
    from opentelemetry.util.types import AttributeValue

SetSpanAttributesFn: TypeAlias = Callable[[Mapping[str, Any]], None]
ChatCompletionChunk: TypeAlias = Union[TextChunk, ToolCallChunk]


@dataclass
class PlaygroundClientCredential:
    """
    Represents a credential for LLM providers.
    """

    env_var_name: str
    value: str


class Dependency:
    """
    Set the module_name to the import name if it is different from the install name
    """

    def __init__(self, name: str, module_name: Optional[str] = None):
        self.name = name
        self.module_name = module_name

    @property
    def import_name(self) -> str:
        return self.module_name or self.name


class KeyedSingleton:
    _instances: dict[Hashable, "KeyedSingleton"] = {}

    def __new__(cls, *args: Any, **kwargs: Any) -> "KeyedSingleton":
        if "singleton_key" in kwargs:
            singleton_key = kwargs.pop("singleton_key")
        elif args:
            singleton_key = args[0]
            args = args[1:]
        else:
            raise ValueError("singleton_key must be provided")

        instance_key = (cls, singleton_key)
        if instance_key not in cls._instances:
            instance = super().__new__(cls)
            cls._instances[instance_key] = instance
        return cls._instances[instance_key]


class PlaygroundRateLimiter(RateLimiter, KeyedSingleton):
    """
    A rate rate limiter class that will be instantiated once per `singleton_key`.
    """

    def __init__(self, singleton_key: Hashable, rate_limit_error: Optional[type[BaseException]]):
        super().__init__(
            rate_limit_error=rate_limit_error,
            max_rate_limit_retries=3,
            initial_per_second_request_rate=1.0,
            maximum_per_second_request_rate=3.0,
            enforcement_window_minutes=0.05,
            rate_reduction_factor=0.5,
            rate_increase_factor=0.01,
            cooldown_seconds=5,
            verbose=False,
        )

    # TODO: update the rate limiter class in phoenix.evals to support decorated sync functions
    def _alimit(
        self, fn: Callable[ParameterSpec, GenericType]
    ) -> AsyncCallable[ParameterSpec, GenericType]:
        @wraps(fn)
        async def wrapper(*args: Any, **kwargs: Any) -> GenericType:
            self._initialize_async_primitives()
            assert self._rate_limit_handling_lock is not None and isinstance(
                self._rate_limit_handling_lock, asyncio.Lock
            )
            assert self._rate_limit_handling is not None and isinstance(
                self._rate_limit_handling, asyncio.Event
            )
            try:
                try:
                    await asyncio.wait_for(self._rate_limit_handling.wait(), 120)
                except asyncio.TimeoutError:
                    self._rate_limit_handling.set()  # Set the event as a failsafe
                await self._throttler.async_wait_until_ready()
                request_start_time = time.time()
                maybe_coroutine = fn(*args, **kwargs)
                if inspect.isawaitable(maybe_coroutine):
                    return await maybe_coroutine  # type: ignore[no-any-return]
                else:
                    return maybe_coroutine
            except self._rate_limit_error:
                async with self._rate_limit_handling_lock:
                    self._rate_limit_handling.clear()  # prevent new requests from starting
                    self._throttler.on_rate_limit_error(request_start_time, verbose=self._verbose)
                    try:
                        for _attempt in range(self._max_rate_limit_retries):
                            try:
                                request_start_time = time.time()
                                await self._throttler.async_wait_until_ready()
                                maybe_coroutine = fn(*args, **kwargs)
                                if inspect.isawaitable(maybe_coroutine):
                                    return await maybe_coroutine  # type: ignore[no-any-return]
                                else:
                                    return maybe_coroutine
                            except self._rate_limit_error:
                                self._throttler.on_rate_limit_error(
                                    request_start_time, verbose=self._verbose
                                )
                                continue
                    finally:
                        self._rate_limit_handling.set()  # allow new requests to start
            raise RateLimitError(f"Exceeded max ({self._max_rate_limit_retries}) retries")

        return wrapper


class PlaygroundStreamingClient(ABC):
    def __init__(
        self,
        model: GenerativeModelInput,
        credentials: Optional[list[PlaygroundClientCredential]] = None,
    ) -> None:
        self._attributes: dict[str, AttributeValue] = dict()
        self._credentials = credentials or []

    @classmethod
    @abstractmethod
    def dependencies(cls) -> list[Dependency]:
        # A list of dependencies this client needs to run
        ...

    @classmethod
    @abstractmethod
    def supported_invocation_parameters(cls) -> list[InvocationParameter]: ...

    @abstractmethod
    async def chat_completion_create(
        self,
        messages: list[
            tuple[ChatCompletionMessageRole, str, Optional[str], Optional[list[JSONScalarType]]]
        ],
        tools: list[JSONScalarType],
        **invocation_parameters: Any,
    ) -> AsyncIterator[ChatCompletionChunk]:
        # a yield statement is needed to satisfy the type-checker
        # https://mypy.readthedocs.io/en/stable/more_types.html#asynchronous-iterators
        yield TextChunk(content="")

    @classmethod
    def construct_invocation_parameters(
        cls, invocation_parameters: list[InvocationParameterInput]
    ) -> dict[str, Any]:
        supported_params = cls.supported_invocation_parameters()
        params = {param.invocation_name: param for param in supported_params}

        formatted_invocation_parameters = dict()

        for param_input in invocation_parameters:
            invocation_name = param_input.invocation_name
            if invocation_name not in params:
                raise ValueError(f"Unsupported invocation parameter: {invocation_name}")

            param_def = params[invocation_name]
            value = extract_parameter(param_def, param_input)
            if value is not UNSET:
                formatted_invocation_parameters[invocation_name] = value
        validate_invocation_parameters(supported_params, formatted_invocation_parameters)
        return formatted_invocation_parameters

    @classmethod
    def dependencies_are_installed(cls) -> bool:
        try:
            for dependency in cls.dependencies():
                import_name = dependency.import_name
                if importlib.util.find_spec(import_name) is None:
                    return False
            return True
        except ValueError:
            # happens in some cases if the spec is None
            return False

    @property
    def attributes(self) -> dict[str, Any]:
        return self._attributes


class OpenAIBaseStreamingClient(PlaygroundStreamingClient):
    def __init__(
        self,
        *,
        client: Union["AsyncOpenAI", "AsyncAzureOpenAI"],
        model: GenerativeModelInput,
        credentials: Optional[list[PlaygroundClientCredential]] = None,
    ) -> None:
        from openai import RateLimitError as OpenAIRateLimitError

        super().__init__(model=model, credentials=credentials)
        self.client = client
        self.model_name = model.name
        self.rate_limiter = PlaygroundRateLimiter(model.provider_key, OpenAIRateLimitError)
        self.client._client = _HttpxClient(self.client._client, self._attributes)

    @classmethod
    def dependencies(cls) -> list[Dependency]:
        return [Dependency(name="openai")]

    @classmethod
    def supported_invocation_parameters(cls) -> list[InvocationParameter]:
        return [
            BoundedFloatInvocationParameter(
                invocation_name="temperature",
                canonical_name=CanonicalParameterName.TEMPERATURE,
                label="Temperature",
                default_value=1.0,
                min_value=0.0,
                max_value=2.0,
            ),
            IntInvocationParameter(
                invocation_name="max_tokens",
                canonical_name=CanonicalParameterName.MAX_COMPLETION_TOKENS,
                label="Max Tokens",
            ),
            BoundedFloatInvocationParameter(
                invocation_name="frequency_penalty",
                label="Frequency Penalty",
                default_value=0.0,
                min_value=-2.0,
                max_value=2.0,
            ),
            BoundedFloatInvocationParameter(
                invocation_name="presence_penalty",
                label="Presence Penalty",
                default_value=0.0,
                min_value=-2.0,
                max_value=2.0,
            ),
            StringListInvocationParameter(
                invocation_name="stop",
                canonical_name=CanonicalParameterName.STOP_SEQUENCES,
                label="Stop Sequences",
            ),
            BoundedFloatInvocationParameter(
                invocation_name="top_p",
                canonical_name=CanonicalParameterName.TOP_P,
                label="Top P",
                min_value=0.0,
                max_value=1.0,
            ),
            IntInvocationParameter(
                invocation_name="seed",
                canonical_name=CanonicalParameterName.RANDOM_SEED,
                label="Seed",
            ),
            JSONInvocationParameter(
                invocation_name="tool_choice",
                label="Tool Choice",
                canonical_name=CanonicalParameterName.TOOL_CHOICE,
            ),
            JSONInvocationParameter(
                invocation_name="response_format",
                label="Response Format",
                canonical_name=CanonicalParameterName.RESPONSE_FORMAT,
            ),
            JSONInvocationParameter(
                invocation_name="extra_body",
                label="Extra Body",
            ),
        ]

    async def chat_completion_create(
        self,
        messages: list[
            tuple[ChatCompletionMessageRole, str, Optional[str], Optional[list[JSONScalarType]]]
        ],
        tools: list[JSONScalarType],
        **invocation_parameters: Any,
    ) -> AsyncIterator[ChatCompletionChunk]:
        from openai import NOT_GIVEN
        from openai.types.chat import ChatCompletionStreamOptionsParam

        # Convert standard messages to OpenAI messages
        openai_messages = []
        for message in messages:
            openai_message = self.to_openai_chat_completion_param(*message)
            if openai_message is not None:
                openai_messages.append(openai_message)
        tool_call_ids: dict[int, str] = {}
        token_usage: Optional["CompletionUsage"] = None
        throttled_create = self.rate_limiter._alimit(self.client.chat.completions.create)
        async for chunk in await throttled_create(
            messages=openai_messages,
            model=self.model_name,
            stream=True,
            stream_options=ChatCompletionStreamOptionsParam(include_usage=True),
            tools=tools or NOT_GIVEN,
            **invocation_parameters,
        ):
            if (usage := chunk.usage) is not None:
                token_usage = usage
            if not chunk.choices:
                # for Azure, initial chunk contains the content filter
                continue
            choice = chunk.choices[0]
            delta = choice.delta
            if choice.finish_reason is None:
                if isinstance(chunk_content := delta.content, str):
                    text_chunk = TextChunk(content=chunk_content)
                    yield text_chunk
                if (tool_calls := delta.tool_calls) is not None:
                    for tool_call_index, tool_call in enumerate(tool_calls):
                        tool_call_id = (
                            tool_call.id
                            if tool_call.id is not None
                            else tool_call_ids[tool_call_index]
                        )
                        tool_call_ids[tool_call_index] = tool_call_id
                        if (function := tool_call.function) is not None:
                            tool_call_chunk = ToolCallChunk(
                                id=tool_call_id,
                                function=FunctionCallChunk(
                                    name=function.name or "",
                                    arguments=function.arguments or "",
                                ),
                            )
                            yield tool_call_chunk
        if token_usage is not None:
            self._attributes.update(dict(self._llm_token_counts(token_usage)))

    def to_openai_chat_completion_param(
        self,
        role: ChatCompletionMessageRole,
        content: JSONScalarType,
        tool_call_id: Optional[str] = None,
        tool_calls: Optional[list[JSONScalarType]] = None,
    ) -> Optional["ChatCompletionMessageParam"]:
        from openai.types.chat import (
            ChatCompletionAssistantMessageParam,
            ChatCompletionSystemMessageParam,
            ChatCompletionToolMessageParam,
            ChatCompletionUserMessageParam,
        )

        if role is ChatCompletionMessageRole.USER:
            return ChatCompletionUserMessageParam(
                {
                    "content": content,
                    "role": "user",
                }
            )
        if role is ChatCompletionMessageRole.SYSTEM:
            return ChatCompletionSystemMessageParam(
                {
                    "content": content,
                    "role": "system",
                }
            )
        if role is ChatCompletionMessageRole.AI:
            if tool_calls is None:
                return ChatCompletionAssistantMessageParam(
                    {
                        "content": content,
                        "role": "assistant",
                    }
                )
            else:
                return ChatCompletionAssistantMessageParam(
                    {
                        "content": content,
                        "role": "assistant",
                        "tool_calls": [
                            self.to_openai_tool_call_param(tool_call) for tool_call in tool_calls
                        ],
                    }
                )
        if role is ChatCompletionMessageRole.TOOL:
            if tool_call_id is None:
                raise ValueError("tool_call_id is required for tool messages")
            return ChatCompletionToolMessageParam(
                {"content": content, "role": "tool", "tool_call_id": tool_call_id}
            )
        assert_never(role)

    def to_openai_tool_call_param(
        self,
        tool_call: JSONScalarType,
    ) -> "ChatCompletionMessageToolCallParam":
        from openai.types.chat import ChatCompletionMessageToolCallParam

        return ChatCompletionMessageToolCallParam(
            id=tool_call.get("id", ""),
            function={
                "name": tool_call.get("function", {}).get("name", ""),
                "arguments": safe_json_dumps(tool_call.get("function", {}).get("arguments", "")),
            },
            type="function",
        )

    @staticmethod
    def _llm_token_counts(usage: "CompletionUsage") -> Iterator[tuple[str, Any]]:
        yield LLM_TOKEN_COUNT_PROMPT, usage.prompt_tokens
        yield LLM_TOKEN_COUNT_COMPLETION, usage.completion_tokens
        yield LLM_TOKEN_COUNT_TOTAL, usage.total_tokens

        if hasattr(usage, "prompt_tokens_details") and usage.prompt_tokens_details is not None:
            prompt_details = usage.prompt_tokens_details
            if (
                hasattr(prompt_details, "cached_tokens")
                and prompt_details.cached_tokens is not None
            ):
                yield LLM_TOKEN_COUNT_PROMPT_DETAILS_CACHE_READ, prompt_details.cached_tokens
            if hasattr(prompt_details, "audio_tokens") and prompt_details.audio_tokens is not None:
                yield LLM_TOKEN_COUNT_PROMPT_DETAILS_AUDIO, prompt_details.audio_tokens

        if (
            hasattr(usage, "completion_tokens_details")
            and usage.completion_tokens_details is not None
        ):
            completion_details = usage.completion_tokens_details
            if (
                hasattr(completion_details, "reasoning_tokens")
                and completion_details.reasoning_tokens is not None
            ):
                yield (
                    LLM_TOKEN_COUNT_COMPLETION_DETAILS_REASONING,
                    completion_details.reasoning_tokens,
                )
            if (
                hasattr(completion_details, "audio_tokens")
                and completion_details.audio_tokens is not None
            ):
                yield LLM_TOKEN_COUNT_COMPLETION_DETAILS_AUDIO, completion_details.audio_tokens


def _get_credential_value(
    credentials: Optional[list[PlaygroundClientCredential]], env_var_name: str
) -> Optional[str]:
    """Helper function to extract credential value from credentials list."""
    if not credentials:
        return None
    return next(
        (credential.value for credential in credentials if credential.env_var_name == env_var_name),
        None,
    )


def _require_credential(
    credentials: Optional[list[PlaygroundClientCredential]], env_var_name: str, provider_name: str
) -> str:
    """Helper function to require a credential value, raising an exception if not found."""
    value = _get_credential_value(credentials, env_var_name)
    if value is None:
        raise BadRequest(f"Missing required credential '{env_var_name}' for {provider_name}")
    return value


@register_llm_client(
    provider_key=GenerativeProviderKey.DEEPSEEK,
    model_names=[
        PROVIDER_DEFAULT,
        "deepseek-chat",
        "deepseek-reasoner",
    ],
)
class DeepSeekStreamingClient(OpenAIBaseStreamingClient):
    def __init__(
        self,
        model: GenerativeModelInput,
        credentials: Optional[list[PlaygroundClientCredential]] = None,
    ) -> None:
        from openai import AsyncOpenAI

        base_url = model.base_url or getenv("DEEPSEEK_BASE_URL")

        # Try to get API key from credentials first, then fallback to env
        api_key = _get_credential_value(credentials, "DEEPSEEK_API_KEY") or getenv(
            "DEEPSEEK_API_KEY"
        )

        if not api_key:
            if not base_url:
                raise BadRequest("An API key is required for DeepSeek models")
            api_key = "sk-fake-api-key"

        client = AsyncOpenAI(
            api_key=api_key,
            base_url=base_url or "https://api.deepseek.com",
            default_headers=model.custom_headers or None,
        )
        super().__init__(client=client, model=model, credentials=credentials)
        # DeepSeek uses OpenAI-compatible API but we'll track it as a separate provider
        # Adding a custom "deepseek" provider value to make it distinguishable in traces
        self._attributes[LLM_PROVIDER] = "deepseek"
        self._attributes[LLM_SYSTEM] = OpenInferenceLLMSystemValues.OPENAI.value


@register_llm_client(
    provider_key=GenerativeProviderKey.XAI,
    model_names=[
        PROVIDER_DEFAULT,
        "grok-3",
        "grok-3-fast",
        "grok-3-mini",
        "grok-3-mini-fast",
        "grok-2-1212",
        "grok-2-vision-1212",
    ],
)
class XAIStreamingClient(OpenAIBaseStreamingClient):
    def __init__(
        self,
        model: GenerativeModelInput,
        credentials: Optional[list[PlaygroundClientCredential]] = None,
    ) -> None:
        from openai import AsyncOpenAI

        base_url = model.base_url or getenv("XAI_BASE_URL")

        # Try to get API key from credentials first, then fallback to env
        api_key = _get_credential_value(credentials, "XAI_API_KEY") or getenv("XAI_API_KEY")

        if not api_key:
            if not base_url:
                raise BadRequest("An API key is required for xAI models")
            api_key = "sk-fake-api-key"

        client = AsyncOpenAI(
            api_key=api_key,
            base_url=base_url or "https://api.x.ai/v1",
            default_headers=model.custom_headers or None,
        )
        super().__init__(client=client, model=model, credentials=credentials)
        # xAI uses OpenAI-compatible API but we'll track it as a separate provider
        # Adding a custom "xai" provider value to make it distinguishable in traces
        self._attributes[LLM_PROVIDER] = "xai"
        self._attributes[LLM_SYSTEM] = OpenInferenceLLMSystemValues.OPENAI.value


@register_llm_client(
    provider_key=GenerativeProviderKey.OLLAMA,
    model_names=[
        PROVIDER_DEFAULT,
        "llama3.3",
        "llama3.2",
        "llama3.1",
        "llama3",
        "llama2",
        "mistral",
        "mixtral",
        "codellama",
        "phi3",
        "qwen2.5",
        "gemma2",
    ],
)
class OllamaStreamingClient(OpenAIBaseStreamingClient):
    def __init__(
        self,
        model: GenerativeModelInput,
        credentials: Optional[list[PlaygroundClientCredential]] = None,
    ) -> None:
        from openai import AsyncOpenAI

        base_url = model.base_url or getenv("OLLAMA_BASE_URL")
        if not base_url:
            raise BadRequest("An Ollama base URL is required for Ollama models")
        api_key = "ollama"
        client = AsyncOpenAI(
            api_key=api_key,
            base_url=base_url,
            default_headers=model.custom_headers or None,
        )
        super().__init__(client=client, model=model, credentials=credentials)
        # Ollama uses OpenAI-compatible API but we'll track it as a separate provider
        # Adding a custom "ollama" provider value to make it distinguishable in traces
        self._attributes[LLM_PROVIDER] = "ollama"
        self._attributes[LLM_SYSTEM] = OpenInferenceLLMSystemValues.OPENAI.value


@register_llm_client(
    provider_key=GenerativeProviderKey.AWS,
    model_names=[
        PROVIDER_DEFAULT,
        "anthropic.claude-opus-4-6-v1",
        "anthropic.claude-opus-4-5-20251101-v1:0",
        "anthropic.claude-sonnet-4-5-20250929-v1:0",
        "anthropic.claude-haiku-4-5-20251001-v1:0",
        "anthropic.claude-opus-4-1-20250805-v1:0",
        "anthropic.claude-opus-4-20250514-v1:0",
        "anthropic.claude-sonnet-4-20250514-v1:0",
        "anthropic.claude-3-7-sonnet-20250219-v1:0",
        "anthropic.claude-3-5-sonnet-20241022-v2:0",
        "anthropic.claude-3-5-sonnet-20240620-v1:0",
        "anthropic.claude-3-5-haiku-20241022-v1:0",
        "anthropic.claude-3-haiku-20240307-v1:0",
        "amazon.titan-embed-text-v2:0",
        "amazon.nova-pro-v1:0",
        "amazon.nova-premier-v1:0:8k",
        "amazon.nova-premier-v1:0:20k",
        "amazon.nova-premier-v1:0:1000k",
        "amazon.nova-premier-v1:0:mm",
        "amazon.nova-premier-v1:0",
        "amazon.nova-lite-v1:0",
        "amazon.nova-micro-v1:0",
        "deepseek.r1-v1:0",
        "mistral.pixtral-large-2502-v1:0",
        "meta.llama3-1-8b-instruct-v1:0:128k",
        "meta.llama3-1-8b-instruct-v1:0",
        "meta.llama3-1-70b-instruct-v1:0:128k",
        "meta.llama3-1-70b-instruct-v1:0",
        "meta.llama3-1-405b-instruct-v1:0",
        "meta.llama3-2-11b-instruct-v1:0",
        "meta.llama3-2-90b-instruct-v1:0",
        "meta.llama3-2-1b-instruct-v1:0",
        "meta.llama3-2-3b-instruct-v1:0",
        "meta.llama3-3-70b-instruct-v1:0",
        "meta.llama4-scout-17b-instruct-v1:0",
        "meta.llama4-maverick-17b-instruct-v1:0",
    ],
)
class BedrockStreamingClient(PlaygroundStreamingClient):
    def __init__(
        self,
        model: GenerativeModelInput,
        credentials: Optional[list[PlaygroundClientCredential]] = None,
    ) -> None:
        import boto3  # type: ignore[import-untyped]

        super().__init__(model=model, credentials=credentials)
        region = model.region or "us-east-1"
        self.api = "converse"
        custom_headers = model.custom_headers
        aws_access_key_id = _get_credential_value(credentials, "AWS_ACCESS_KEY_ID") or getenv(
            "AWS_ACCESS_KEY_ID"
        )
        aws_secret_access_key = _get_credential_value(
            credentials, "AWS_SECRET_ACCESS_KEY"
        ) or getenv("AWS_SECRET_ACCESS_KEY")
        aws_session_token = _get_credential_value(credentials, "AWS_SESSION_TOKEN") or getenv(
            "AWS_SESSION_TOKEN"
        )
        self.model_name = model.name
        session = boto3.Session(
            region_name=region,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token,
        )
        client = session.client(service_name="bedrock-runtime")

        # Add custom headers support via boto3 event system
        if custom_headers:

            def add_custom_headers(request: "AWSPreparedRequest", **kwargs: Any) -> None:
                request.headers.update(custom_headers)

            client.meta.events.register("before-send.*", add_custom_headers)

        self.client = client
        self._attributes[LLM_PROVIDER] = "aws"
        self._attributes[LLM_SYSTEM] = "aws"

    @staticmethod
    def _setup_custom_headers(client: Any, custom_headers: Mapping[str, str]) -> None:
        """Setup custom headers using boto3's event system."""
        if not custom_headers:
            return

    @classmethod
    def dependencies(cls) -> list[Dependency]:
        return [Dependency(name="boto3")]

    @classmethod
    def supported_invocation_parameters(cls) -> list[InvocationParameter]:
        return [
            IntInvocationParameter(
                invocation_name="max_tokens",
                canonical_name=CanonicalParameterName.MAX_COMPLETION_TOKENS,
                label="Max Tokens",
                default_value=1024,
            ),
            BoundedFloatInvocationParameter(
                invocation_name="temperature",
                canonical_name=CanonicalParameterName.TEMPERATURE,
                label="Temperature",
                default_value=1.0,
                min_value=0.0,
                max_value=1.0,
            ),
            BoundedFloatInvocationParameter(
                invocation_name="top_p",
                canonical_name=CanonicalParameterName.TOP_P,
                label="Top P",
                min_value=0.0,
                max_value=1.0,
            ),
            JSONInvocationParameter(
                invocation_name="tool_choice",
                label="Tool Choice",
                canonical_name=CanonicalParameterName.TOOL_CHOICE,
            ),
        ]

    async def chat_completion_create(
        self,
        messages: list[
            tuple[ChatCompletionMessageRole, str, Optional[str], Optional[list[JSONScalarType]]]
        ],
        tools: list[JSONScalarType],
        **invocation_parameters: Any,
    ) -> AsyncIterator[ChatCompletionChunk]:
        if self.api == "invoke":
            async for chunk in self._handle_invoke_api(messages, tools, invocation_parameters):
                yield chunk
        else:
            async for chunk in self._handle_converse_api(messages, tools, invocation_parameters):
                yield chunk

    async def _handle_converse_api(
        self,
        messages: list[
            tuple[ChatCompletionMessageRole, str, Optional[str], Optional[list[JSONScalarType]]]
        ],
        tools: list[JSONScalarType],
        invocation_parameters: dict[str, Any],
    ) -> AsyncIterator[ChatCompletionChunk]:
        """
        Handle the converse API.
        """
        # Build messages in Converse API format
        converse_messages = self._build_converse_messages(messages)

        inference_config = {}
        if (
            "max_tokens" in invocation_parameters
            and invocation_parameters["max_tokens"] is not None
        ):
            inference_config["maxTokens"] = invocation_parameters["max_tokens"]
        if (
            "temperature" in invocation_parameters
            and invocation_parameters["temperature"] is not None
        ):
            inference_config["temperature"] = invocation_parameters["temperature"]
        if "top_p" in invocation_parameters and invocation_parameters["top_p"] is not None:
            inference_config["topP"] = invocation_parameters["top_p"]

        # Build the request parameters for Converse API
        converse_params: dict[str, Any] = {
            "modelId": self.model_name,
            "messages": converse_messages,
            "inferenceConfig": inference_config,
        }

        # Add system prompt if available
        system_prompt = self._extract_system_prompt(messages)
        if system_prompt:
            converse_params["system"] = [{"text": system_prompt}]

        # Add tools if provided
        if tools:
            converse_params["toolConfig"] = {"tools": tools}
            if (
                "tool_choice" in invocation_parameters
                and invocation_parameters["tool_choice"]["type"] != "none"
            ):
                converse_params["toolConfig"]["toolChoice"] = {}

                if invocation_parameters["tool_choice"]["type"] == "auto":
                    converse_params["toolConfig"]["toolChoice"]["auto"] = {}
                elif invocation_parameters["tool_choice"]["type"] == "any":
                    converse_params["toolConfig"]["toolChoice"]["any"] = {}
                else:
                    converse_params["toolConfig"]["toolChoice"]["tool"] = {
                        "name": invocation_parameters["tool_choice"]["name"],
                    }

        # Make the streaming API call
        response = self.client.converse_stream(**converse_params)

        # Track active tool calls
        active_tool_calls = {}  # contentBlockIndex -> {id, name, arguments_buffer}

        # Process the event stream
        event_stream = response.get("stream")

        for event in event_stream:
            # Handle content block start events
            if "contentBlockStart" in event:
                content_block_start = event["contentBlockStart"]
                start_event = content_block_start.get("start", {})
                block_index = content_block_start.get(
                    "contentBlockIndex", 0
                )  # Get the actual index

                if "toolUse" in start_event:
                    tool_use = start_event["toolUse"]
                    active_tool_calls[block_index] = {  # Use the actual block index
                        "id": tool_use.get("toolUseId"),
                        "name": tool_use.get("name"),
                        "arguments_buffer": "",
                    }

                    # Yield initial tool call chunk
                    yield ToolCallChunk(
                        id=tool_use.get("toolUseId"),
                        function=FunctionCallChunk(
                            name=tool_use.get("name"),
                            arguments="",
                        ),
                    )

            # Handle content block delta events
            elif "contentBlockDelta" in event:
                content_delta = event["contentBlockDelta"]
                delta = content_delta.get("delta", {})
                delta_index = content_delta.get("contentBlockIndex", 0)

                # Handle text delta
                if "text" in delta:
                    yield TextChunk(content=delta["text"])

                # Handle tool use delta
                elif "toolUse" in delta:
                    tool_delta = delta["toolUse"]
                    if "input" in tool_delta and delta_index in active_tool_calls:
                        # Accumulate tool arguments
                        json_chunk = tool_delta["input"]
                        active_tool_calls[delta_index]["arguments_buffer"] += json_chunk

                        # Yield incremental argument update
                        yield ToolCallChunk(
                            id=active_tool_calls[delta_index]["id"],
                            function=FunctionCallChunk(
                                name=active_tool_calls[delta_index]["name"],
                                arguments=json_chunk,
                            ),
                        )

            # Handle content block stop events
            elif "contentBlockStop" in event:
                stop_index = event["contentBlockStop"].get("contentBlockIndex", 0)
                if stop_index in active_tool_calls:
                    del active_tool_calls[stop_index]

            elif "metadata" in event:
                self._attributes.update(
                    {
                        LLM_TOKEN_COUNT_PROMPT: event.get("metadata")
                        .get("usage", {})
                        .get("inputTokens", 0)
                    }
                )

                self._attributes.update(
                    {
                        LLM_TOKEN_COUNT_COMPLETION: event.get("metadata")
                        .get("usage", {})
                        .get("outputTokens", 0)
                    }
                )

                self._attributes.update(
                    {
                        LLM_TOKEN_COUNT_TOTAL: event.get("metadata")
                        .get("usage", {})
                        .get("totalTokens", 0)
                    }
                )

    async def _handle_invoke_api(
        self,
        messages: list[
            tuple[ChatCompletionMessageRole, str, Optional[str], Optional[list[JSONScalarType]]]
        ],
        tools: list[JSONScalarType],
        invocation_parameters: dict[str, Any],
    ) -> AsyncIterator[ChatCompletionChunk]:
        if "anthropic" not in self.model_name:
            raise ValueError("Invoke API is only supported for Anthropic models")

        bedrock_messages, system_prompt = self._build_bedrock_messages(messages)
        bedrock_params = {
            "anthropic_version": "bedrock-2023-05-31",
            "messages": bedrock_messages,
            "system": system_prompt,
            "tools": tools,
        }

        if (
            "max_tokens" in invocation_parameters
            and invocation_parameters["max_tokens"] is not None
        ):
            bedrock_params["max_tokens"] = invocation_parameters["max_tokens"]
        if (
            "temperature" in invocation_parameters
            and invocation_parameters["temperature"] is not None
        ):
            bedrock_params["temperature"] = invocation_parameters["temperature"]
        if "top_p" in invocation_parameters and invocation_parameters["top_p"] is not None:
            bedrock_params["top_p"] = invocation_parameters["top_p"]

        response = self.client.invoke_model_with_response_stream(
            modelId=self.model_name,
            contentType="application/json",
            accept="application/json",
            body=json.dumps(bedrock_params),
            trace="ENABLED_FULL",
        )

        # The response['body'] is an EventStream object
        event_stream = response["body"]

        # Track active tool calls and their accumulating arguments
        active_tool_calls: dict[int, dict[str, Any]] = {}  # index -> {id, name, arguments_buffer}

        for event in event_stream:
            if "chunk" in event:
                chunk_data = json.loads(event["chunk"]["bytes"].decode("utf-8"))

                # Handle text content
                if chunk_data.get("type") == "content_block_delta":
                    delta = chunk_data.get("delta", {})
                    index = chunk_data.get("index", 0)

                    if delta.get("type") == "text_delta" and "text" in delta:
                        yield TextChunk(content=delta["text"])

                    elif delta.get("type") == "input_json_delta":
                        # Accumulate tool arguments
                        if index in active_tool_calls:
                            active_tool_calls[index]["arguments_buffer"] += delta.get(
                                "partial_json", ""
                            )
                            # Yield incremental argument update
                            yield ToolCallChunk(
                                id=active_tool_calls[index]["id"],
                                function=FunctionCallChunk(
                                    name=active_tool_calls[index]["name"],
                                    arguments=delta.get("partial_json", ""),
                                ),
                            )

                # Handle tool call start
                elif chunk_data.get("type") == "content_block_start":
                    content_block = chunk_data.get("content_block", {})
                    index = chunk_data.get("index", 0)

                    if content_block.get("type") == "tool_use":
                        # Initialize tool call tracking
                        active_tool_calls[index] = {
                            "id": content_block.get("id"),
                            "name": content_block.get("name"),
                            "arguments_buffer": "",
                        }

                        # Yield initial tool call chunk
                        yield ToolCallChunk(
                            id=content_block.get("id"),
                            function=FunctionCallChunk(
                                name=content_block.get("name"),
                                arguments="",  # Start with empty, will be filled by deltas
                            ),
                        )

                # Handle content block stop (tool call complete)
                elif chunk_data.get("type") == "content_block_stop":
                    index = chunk_data.get("index", 0)
                    if index in active_tool_calls:
                        # Tool call is complete, clean up
                        del active_tool_calls[index]

                elif chunk_data.get("type") == "message_stop":
                    self._attributes.update(
                        {
                            LLM_TOKEN_COUNT_COMPLETION: chunk_data.get(
                                "amazon-bedrock-invocationMetrics", {}
                            ).get("outputTokenCount", 0)
                        }
                    )

                    self._attributes.update(
                        {
                            LLM_TOKEN_COUNT_PROMPT: chunk_data.get(
                                "amazon-bedrock-invocationMetrics", {}
                            ).get("inputTokenCount", 0)
                        }
                    )

    def _build_bedrock_messages(
        self,
        messages: list[
            tuple[ChatCompletionMessageRole, str, Optional[str], Optional[list[JSONScalarType]]]
        ],
    ) -> tuple[list[dict[str, Any]], str]:
        bedrock_messages = []
        system_prompt = ""
        for role, content, _, _ in messages:
            if role == ChatCompletionMessageRole.USER:
                bedrock_messages.append(
                    {
                        "role": "user",
                        "content": content,
                    }
                )
            elif role == ChatCompletionMessageRole.AI:
                bedrock_messages.append(
                    {
                        "role": "assistant",
                        "content": content,
                    }
                )
            elif role == ChatCompletionMessageRole.SYSTEM:
                system_prompt += content + "\n"
        return bedrock_messages, system_prompt

    def _extract_system_prompt(
        self,
        messages: list[
            tuple[ChatCompletionMessageRole, str, Optional[str], Optional[list[JSONScalarType]]]
        ],
    ) -> str:
        """Extract system prompt from messages."""
        system_prompts = []
        for role, content, _, _ in messages:
            if role == ChatCompletionMessageRole.SYSTEM:
                system_prompts.append(content)
        return "\n".join(system_prompts)

    def _build_converse_messages(
        self,
        messages: list[
            tuple[ChatCompletionMessageRole, str, Optional[str], Optional[list[JSONScalarType]]]
        ],
    ) -> list[dict[str, Any]]:
        """Convert messages to Converse API format."""
        converse_messages: list[dict[str, Any]] = []
        for role, content, _id, tool_calls in messages:
            if role == ChatCompletionMessageRole.USER:
                converse_messages.append({"role": "user", "content": [{"text": content}]})
            elif role == ChatCompletionMessageRole.TOOL:
                converse_messages.append(
                    {
                        "role": "user",
                        "content": [
                            {
                                "toolResult": {
                                    "toolUseId": _id,
                                    "content": [{"json": json.loads(content)}],
                                }
                            }
                        ],
                    }
                )

            elif role == ChatCompletionMessageRole.AI:
                # Handle assistant messages with potential tool calls
                message: dict[str, Any] = {"role": "assistant", "content": []}
                if content:
                    message["content"].append({"text": content})
                if tool_calls:
                    for tool_call in tool_calls:
                        message["content"].append(tool_call)
                converse_messages.append(message)
        return converse_messages


@register_llm_client(
    provider_key=GenerativeProviderKey.OPENAI,
    model_names=[
        PROVIDER_DEFAULT,
        "gpt-4.1",
        "gpt-4.1-mini",
        "gpt-4.1-nano",
        "gpt-4.1-2025-04-14",
        "gpt-4.1-mini-2025-04-14",
        "gpt-4.1-nano-2025-04-14",
        "gpt-4o",
        "gpt-4o-2024-11-20",
        "gpt-4o-2024-08-06",
        "gpt-4o-2024-05-13",
        "chatgpt-4o-latest",
        "gpt-4o-mini",
        "gpt-4o-mini-2024-07-18",
        "gpt-4-turbo",
        "gpt-4-turbo-2024-04-09",
        "gpt-4-turbo-preview",
        "gpt-4-0125-preview",
        "gpt-4-1106-preview",
        "gpt-4",
        "gpt-4-0613",
        "gpt-3.5-turbo-0125",
        "gpt-3.5-turbo",
        "gpt-3.5-turbo-1106",
        # preview models
        "gpt-4.5-preview",
    ],
)
class OpenAIStreamingClient(OpenAIBaseStreamingClient):
    def __init__(
        self,
        model: GenerativeModelInput,
        credentials: Optional[list[PlaygroundClientCredential]] = None,
    ) -> None:
        from openai import AsyncOpenAI

        base_url = model.base_url or getenv("OPENAI_BASE_URL")

        # Try to get API key from credentials first, then fallback to env
        api_key = _get_credential_value(credentials, "OPENAI_API_KEY") or getenv("OPENAI_API_KEY")

        if not api_key:
            if not base_url:
                raise BadRequest("An API key is required for OpenAI models")
            api_key = "sk-fake-api-key"

        client = AsyncOpenAI(
            api_key=api_key,
            base_url=base_url,
            default_headers=model.custom_headers or None,
            timeout=30,
        )
        super().__init__(client=client, model=model, credentials=credentials)
        self._attributes[LLM_PROVIDER] = OpenInferenceLLMProviderValues.OPENAI.value
        self._attributes[LLM_SYSTEM] = OpenInferenceLLMSystemValues.OPENAI.value


_OPENAI_REASONING_MODELS = [
    "gpt-5.2",
    "gpt-5.2-2025-12-11",
    "gpt-5.2-chat-latest",
    "gpt-5.1",
    "gpt-5.1-2025-11-13",
    "gpt-5.1-chat-latest",
    "gpt-5",
    "gpt-5-mini",
    "gpt-5-nano",
    "gpt-5-chat-latest",
    "o1",
    "o1-pro",
    "o1-2024-12-17",
    "o1-pro-2025-03-19",
    "o1-mini",
    "o1-mini-2024-09-12",
    "o1-preview",
    "o1-preview-2024-09-12",
    "o3",
    "o3-pro",
    "o3-2025-04-16",
    "o3-mini",
    "o3-mini-2025-01-31",
    "o4-mini",
    "o4-mini-2025-04-16",
]


class OpenAIReasoningReasoningModelsMixin:
    """Mixin class for OpenAI-style reasoning model clients (o1, o3 series)."""

    @classmethod
    def supported_invocation_parameters(cls) -> list[InvocationParameter]:
        return [
            StringInvocationParameter(
                invocation_name="reasoning_effort",
                label="Reasoning Effort",
                canonical_name=CanonicalParameterName.REASONING_EFFORT,
            ),
            IntInvocationParameter(
                invocation_name="max_completion_tokens",
                canonical_name=CanonicalParameterName.MAX_COMPLETION_TOKENS,
                label="Max Completion Tokens",
            ),
            IntInvocationParameter(
                invocation_name="seed",
                canonical_name=CanonicalParameterName.RANDOM_SEED,
                label="Seed",
            ),
            JSONInvocationParameter(
                invocation_name="tool_choice",
                label="Tool Choice",
                canonical_name=CanonicalParameterName.TOOL_CHOICE,
            ),
            JSONInvocationParameter(
                invocation_name="response_format",
                label="Response Format",
                canonical_name=CanonicalParameterName.RESPONSE_FORMAT,
            ),
            JSONInvocationParameter(
                invocation_name="extra_body",
                label="Extra Body",
            ),
        ]


@register_llm_client(
    provider_key=GenerativeProviderKey.OPENAI,
    model_names=_OPENAI_REASONING_MODELS,
)
class OpenAIReasoningNonStreamingClient(
    OpenAIReasoningReasoningModelsMixin,
    OpenAIStreamingClient,
):
    def to_openai_chat_completion_param(
        self,
        role: ChatCompletionMessageRole,
        content: JSONScalarType,
        tool_call_id: Optional[str] = None,
        tool_calls: Optional[list[JSONScalarType]] = None,
    ) -> Optional["ChatCompletionMessageParam"]:
        from openai.types.chat import (
            ChatCompletionAssistantMessageParam,
            ChatCompletionDeveloperMessageParam,
            ChatCompletionToolMessageParam,
            ChatCompletionUserMessageParam,
        )

        if role is ChatCompletionMessageRole.USER:
            return ChatCompletionUserMessageParam(
                {
                    "content": content,
                    "role": "user",
                }
            )
        if role is ChatCompletionMessageRole.SYSTEM:
            return ChatCompletionDeveloperMessageParam(
                {
                    "content": content,
                    "role": "developer",
                }
            )
        if role is ChatCompletionMessageRole.AI:
            if tool_calls is None:
                return ChatCompletionAssistantMessageParam(
                    {
                        "content": content,
                        "role": "assistant",
                    }
                )
            else:
                return ChatCompletionAssistantMessageParam(
                    {
                        "content": content,
                        "role": "assistant",
                        "tool_calls": [
                            self.to_openai_tool_call_param(tool_call) for tool_call in tool_calls
                        ],
                    }
                )
        if role is ChatCompletionMessageRole.TOOL:
            if tool_call_id is None:
                raise ValueError("tool_call_id is required for tool messages")
            return ChatCompletionToolMessageParam(
                {"content": content, "role": "tool", "tool_call_id": tool_call_id}
            )
        assert_never(role)


@register_llm_client(
    provider_key=GenerativeProviderKey.AZURE_OPENAI,
    model_names=[
        PROVIDER_DEFAULT,
    ],
)
class AzureOpenAIStreamingClient(OpenAIBaseStreamingClient):
    def __init__(
        self,
        model: GenerativeModelInput,
        credentials: Optional[list[PlaygroundClientCredential]] = None,
    ):
        from openai import AsyncAzureOpenAI

        if not (endpoint := model.endpoint or getenv("AZURE_OPENAI_ENDPOINT")):
            raise BadRequest("An Azure endpoint is required for Azure OpenAI models")
        if not (api_version := model.api_version or getenv("OPENAI_API_VERSION")):
            raise BadRequest("An OpenAI API version is required for Azure OpenAI models")

        # Try to get API key from credentials first, then fallback to env
        api_key = _get_credential_value(credentials, "AZURE_OPENAI_API_KEY") or getenv(
            "AZURE_OPENAI_API_KEY"
        )

        if api_key:
            client = AsyncAzureOpenAI(
                api_key=api_key,
                azure_endpoint=endpoint,
                api_version=api_version,
                default_headers=model.custom_headers or None,
            )
        else:
            try:
                from azure.identity.aio import DefaultAzureCredential, get_bearer_token_provider
            except ImportError:
                raise BadRequest(
                    "Provide an API key for Azure OpenAI models or use azure-identity, see. e.g. "
                    "https://learn.microsoft.com/en-us/python/api/azure-identity/azure.identity.environmentcredential?view=azure-python"  # noqa: E501
                )

            client = AsyncAzureOpenAI(
                azure_ad_token_provider=get_bearer_token_provider(
                    DefaultAzureCredential(),
                    "https://cognitiveservices.azure.com/.default",
                ),
                azure_endpoint=endpoint,
                api_version=api_version,
                default_headers=model.custom_headers or None,
            )
        super().__init__(client=client, model=model, credentials=credentials)
        self._attributes[LLM_PROVIDER] = OpenInferenceLLMProviderValues.AZURE.value
        self._attributes[LLM_SYSTEM] = OpenInferenceLLMSystemValues.OPENAI.value


@register_llm_client(
    provider_key=GenerativeProviderKey.AZURE_OPENAI,
    model_names=_OPENAI_REASONING_MODELS,
)
class AzureOpenAIReasoningNonStreamingClient(
    OpenAIReasoningReasoningModelsMixin,
    AzureOpenAIStreamingClient,
):
    @override
    async def chat_completion_create(
        self,
        messages: list[
            tuple[ChatCompletionMessageRole, str, Optional[str], Optional[list[JSONScalarType]]]
        ],
        tools: list[JSONScalarType],
        **invocation_parameters: Any,
    ) -> AsyncIterator[ChatCompletionChunk]:
        from openai import NOT_GIVEN

        # Convert standard messages to OpenAI messages
        openai_messages = []
        for message in messages:
            openai_message = self.to_openai_chat_completion_param(*message)
            if openai_message is not None:
                openai_messages.append(openai_message)

        throttled_create = self.rate_limiter._alimit(self.client.chat.completions.create)
        response = await throttled_create(
            messages=openai_messages,
            model=self.model_name,
            stream=False,
            tools=tools or NOT_GIVEN,
            **invocation_parameters,
        )

        if response.usage is not None:
            self._attributes.update(dict(self._llm_token_counts(response.usage)))

        choice = response.choices[0]
        if choice.message.content:
            yield TextChunk(content=choice.message.content)

        if choice.message.tool_calls:
            for tool_call in choice.message.tool_calls:
                yield ToolCallChunk(
                    id=tool_call.id,
                    function=FunctionCallChunk(
                        name=tool_call.function.name,
                        arguments=tool_call.function.arguments,
                    ),
                )

    def to_openai_chat_completion_param(
        self,
        role: ChatCompletionMessageRole,
        content: JSONScalarType,
        tool_call_id: Optional[str] = None,
        tool_calls: Optional[list[JSONScalarType]] = None,
    ) -> Optional["ChatCompletionMessageParam"]:
        from openai.types.chat import (
            ChatCompletionAssistantMessageParam,
            ChatCompletionDeveloperMessageParam,
            ChatCompletionToolMessageParam,
            ChatCompletionUserMessageParam,
        )

        if role is ChatCompletionMessageRole.USER:
            return ChatCompletionUserMessageParam(
                {
                    "content": content,
                    "role": "user",
                }
            )
        if role is ChatCompletionMessageRole.SYSTEM:
            return ChatCompletionDeveloperMessageParam(
                {
                    "content": content,
                    "role": "developer",
                }
            )
        if role is ChatCompletionMessageRole.AI:
            if tool_calls is None:
                return ChatCompletionAssistantMessageParam(
                    {
                        "content": content,
                        "role": "assistant",
                    }
                )
            else:
                return ChatCompletionAssistantMessageParam(
                    {
                        "content": content,
                        "role": "assistant",
                        "tool_calls": [
                            self.to_openai_tool_call_param(tool_call) for tool_call in tool_calls
                        ],
                    }
                )
        if role is ChatCompletionMessageRole.TOOL:
            if tool_call_id is None:
                raise ValueError("tool_call_id is required for tool messages")
            return ChatCompletionToolMessageParam(
                {"content": content, "role": "tool", "tool_call_id": tool_call_id}
            )
        assert_never(role)


@register_llm_client(
    provider_key=GenerativeProviderKey.ANTHROPIC,
    model_names=[
        PROVIDER_DEFAULT,
        "claude-3-5-haiku-latest",
        "claude-3-5-haiku-20241022",
        "claude-3-haiku-20240307",
    ],
)
class AnthropicStreamingClient(PlaygroundStreamingClient):
    def __init__(
        self,
        model: GenerativeModelInput,
        credentials: Optional[list[PlaygroundClientCredential]] = None,
    ) -> None:
        import anthropic

        super().__init__(model=model, credentials=credentials)
        self._attributes[LLM_PROVIDER] = OpenInferenceLLMProviderValues.ANTHROPIC.value
        self._attributes[LLM_SYSTEM] = OpenInferenceLLMSystemValues.ANTHROPIC.value

        # Try to get API key from credentials first, then fallback to env
        api_key = _get_credential_value(credentials, "ANTHROPIC_API_KEY") or getenv(
            "ANTHROPIC_API_KEY"
        )

        if not api_key:
            raise BadRequest("An API key is required for Anthropic models")

        self.client = anthropic.AsyncAnthropic(
            api_key=api_key,
            default_headers=model.custom_headers or None,
        )
        self.model_name = model.name
        self.rate_limiter = PlaygroundRateLimiter(model.provider_key, anthropic.RateLimitError)
        self.client._client = _HttpxClient(self.client._client, self._attributes)

    @classmethod
    def dependencies(cls) -> list[Dependency]:
        return [Dependency(name="anthropic")]

    @classmethod
    def supported_invocation_parameters(cls) -> list[InvocationParameter]:
        return [
            IntInvocationParameter(
                invocation_name="max_tokens",
                canonical_name=CanonicalParameterName.MAX_COMPLETION_TOKENS,
                label="Max Tokens",
                default_value=1024,
                required=True,
            ),
            BoundedFloatInvocationParameter(
                invocation_name="temperature",
                canonical_name=CanonicalParameterName.TEMPERATURE,
                label="Temperature",
                default_value=1.0,
                min_value=0.0,
                max_value=1.0,
            ),
            StringListInvocationParameter(
                invocation_name="stop_sequences",
                canonical_name=CanonicalParameterName.STOP_SEQUENCES,
                label="Stop Sequences",
            ),
            BoundedFloatInvocationParameter(
                invocation_name="top_p",
                canonical_name=CanonicalParameterName.TOP_P,
                label="Top P",
                min_value=0.0,
                max_value=1.0,
            ),
            JSONInvocationParameter(
                invocation_name="tool_choice",
                label="Tool Choice",
                canonical_name=CanonicalParameterName.TOOL_CHOICE,
            ),
        ]

    async def chat_completion_create(
        self,
        messages: list[
            tuple[ChatCompletionMessageRole, str, Optional[str], Optional[list[JSONScalarType]]]
        ],
        tools: list[JSONScalarType],
        **invocation_parameters: Any,
    ) -> AsyncIterator[ChatCompletionChunk]:
        anthropic_messages, system_prompt = self._build_anthropic_messages(messages)
        anthropic_params = {
            "messages": anthropic_messages,
            "model": self.model_name,
            "system": system_prompt,
            "tools": tools,
            **invocation_parameters,
        }
        throttled_stream = self.rate_limiter._alimit(self.client.messages.stream)
        async with await throttled_stream(**anthropic_params) as stream:
            async for event in stream:
                if event.type == "message_start":
                    usage = event.message.usage

                    token_counts: dict[str, Any] = {}
                    if prompt_tokens := (
                        (usage.input_tokens or 0)
                        + (getattr(usage, "cache_creation_input_tokens", 0) or 0)
                        + (getattr(usage, "cache_read_input_tokens", 0) or 0)
                    ):
                        token_counts[LLM_TOKEN_COUNT_PROMPT] = prompt_tokens
                    if cache_creation_tokens := getattr(usage, "cache_creation_input_tokens", None):
                        if cache_creation_tokens is not None:
                            token_counts[LLM_TOKEN_COUNT_PROMPT_DETAILS_CACHE_WRITE] = (
                                cache_creation_tokens
                            )
                    self._attributes.update(token_counts)
                elif event.type == "text":
                    yield TextChunk(content=event.text)
                elif event.type == "message_stop":
                    usage = event.message.usage
                    output_token_counts: dict[str, Any] = {}
                    if usage.output_tokens:
                        output_token_counts[LLM_TOKEN_COUNT_COMPLETION] = usage.output_tokens
                    if cache_read_tokens := getattr(usage, "cache_read_input_tokens", None):
                        if cache_read_tokens is not None:
                            output_token_counts[LLM_TOKEN_COUNT_PROMPT_DETAILS_CACHE_READ] = (
                                cache_read_tokens
                            )
                    self._attributes.update(output_token_counts)
                elif event.type == "content_block_stop" and event.content_block.type == "tool_use":
                    tool_call_chunk = ToolCallChunk(
                        id=event.content_block.id,
                        function=FunctionCallChunk(
                            name=event.content_block.name,
                            arguments=json.dumps(event.content_block.input),
                        ),
                    )
                    yield tool_call_chunk
                elif event.type == "content_block_start":
                    pass
                elif event.type == "content_block_delta":
                    pass
                elif event.type == "message_delta":
                    pass
                elif event.type == "content_block_stop":
                    # non-tool_use case; tool_use already yielded above
                    pass
                elif event.type == "input_json":
                    # Incremental tool-call JSON; we use the complete block at content_block_stop
                    pass
                elif event.type == "citation":
                    pass
                elif event.type == "thinking":
                    pass
                elif event.type == "signature":
                    pass
                elif TYPE_CHECKING:
                    assert_never(event)

    def _build_anthropic_messages(
        self,
        messages: list[tuple[ChatCompletionMessageRole, str, Optional[str], Optional[list[str]]]],
    ) -> tuple[list["MessageParam"], str]:
        anthropic_messages: list["MessageParam"] = []
        system_prompt = ""
        for role, content, _tool_call_id, _tool_calls in messages:
            tool_aware_content = self._anthropic_message_content(content, _tool_calls)
            if role == ChatCompletionMessageRole.USER:
                anthropic_messages.append({"role": "user", "content": tool_aware_content})
            elif role == ChatCompletionMessageRole.AI:
                anthropic_messages.append({"role": "assistant", "content": tool_aware_content})
            elif role == ChatCompletionMessageRole.SYSTEM:
                system_prompt += content + "\n"
            elif role == ChatCompletionMessageRole.TOOL:
                anthropic_messages.append(
                    {
                        "role": "user",
                        "content": [
                            {
                                "type": "tool_result",
                                "tool_use_id": _tool_call_id or "",
                                "content": content or "",
                            }
                        ],
                    }
                )
            else:
                assert_never(role)

        return anthropic_messages, system_prompt

    def _anthropic_message_content(
        self, content: str, tool_calls: Optional[list[JSONScalarType]]
    ) -> Union[str, list[Union["ToolResultBlockParam", "TextBlockParam"]]]:
        if tool_calls:
            # Anthropic combines tool calls and the reasoning text into a single message object
            tool_use_content: list[Union["ToolResultBlockParam", "TextBlockParam"]] = []
            if content:
                tool_use_content.append({"type": "text", "text": content})
            tool_use_content.extend(tool_calls)
            return tool_use_content

        return content


@register_llm_client(
    provider_key=GenerativeProviderKey.ANTHROPIC,
    model_names=[
        "claude-opus-4-6",
        "claude-opus-4-5",
        "claude-opus-4-5-20251101",
        "claude-sonnet-4-5",
        "claude-sonnet-4-5-20250929",
        "claude-haiku-4-5",
        "claude-haiku-4-5-20251001",
        "claude-opus-4-1",
        "claude-opus-4-1-20250805",
        "claude-sonnet-4-0",
        "claude-sonnet-4-20250514",
        "claude-opus-4-0",
        "claude-opus-4-20250514",
        "claude-3-7-sonnet-latest",
        "claude-3-7-sonnet-20250219",
    ],
)
class AnthropicReasoningStreamingClient(AnthropicStreamingClient):
    @classmethod
    def supported_invocation_parameters(cls) -> list[InvocationParameter]:
        invocation_params = super().supported_invocation_parameters()
        invocation_params.append(
            JSONInvocationParameter(
                invocation_name="thinking",
                canonical_name=CanonicalParameterName.ANTHROPIC_EXTENDED_THINKING,
                label="Thinking Budget",
            )
        )
        return invocation_params


@register_llm_client(
    provider_key=GenerativeProviderKey.GOOGLE,
    model_names=[
        PROVIDER_DEFAULT,
        "gemini-2.0-flash-lite",
        "gemini-2.0-flash-001",
        "gemini-2.0-flash-thinking-exp-01-21",
        "gemini-1.5-flash",
        "gemini-1.5-flash-8b",
        "gemini-1.5-pro",
        "gemini-1.0-pro",
    ],
)
class GoogleStreamingClient(PlaygroundStreamingClient):
    def __init__(
        self,
        model: GenerativeModelInput,
        credentials: Optional[list[PlaygroundClientCredential]] = None,
    ) -> None:
        import google.genai as google_genai

        super().__init__(model=model, credentials=credentials)
        self._attributes[LLM_PROVIDER] = OpenInferenceLLMProviderValues.GOOGLE.value
        self._attributes[LLM_SYSTEM] = OpenInferenceLLMSystemValues.VERTEXAI.value

        # Try to get API key from credentials first, then fallback to env
        api_key = (
            _get_credential_value(credentials, "GEMINI_API_KEY")
            or _get_credential_value(credentials, "GOOGLE_API_KEY")
            or getenv("GEMINI_API_KEY")
            or getenv("GOOGLE_API_KEY")
        )

        if not api_key:
            raise BadRequest("An API key is required for Gemini models")

        self.client = google_genai.Client(api_key=api_key)
        self.model_name = model.name

    @classmethod
    def dependencies(cls) -> list[Dependency]:
        return [Dependency(name="google-genai", module_name="google.genai")]

    @classmethod
    def supported_invocation_parameters(cls) -> list[InvocationParameter]:
        return [
            BoundedFloatInvocationParameter(
                invocation_name="temperature",
                canonical_name=CanonicalParameterName.TEMPERATURE,
                label="Temperature",
                default_value=1.0,
                min_value=0.0,
                max_value=2.0,
            ),
            IntInvocationParameter(
                invocation_name="max_output_tokens",
                canonical_name=CanonicalParameterName.MAX_COMPLETION_TOKENS,
                label="Max Output Tokens",
            ),
            StringListInvocationParameter(
                invocation_name="stop_sequences",
                canonical_name=CanonicalParameterName.STOP_SEQUENCES,
                label="Stop Sequences",
            ),
            FloatInvocationParameter(
                invocation_name="presence_penalty",
                label="Presence Penalty",
                default_value=0.0,
            ),
            FloatInvocationParameter(
                invocation_name="frequency_penalty",
                label="Frequency Penalty",
                default_value=0.0,
            ),
            BoundedFloatInvocationParameter(
                invocation_name="top_p",
                canonical_name=CanonicalParameterName.TOP_P,
                label="Top P",
                min_value=0.0,
                max_value=1.0,
            ),
            IntInvocationParameter(
                invocation_name="top_k",
                label="Top K",
            ),
            JSONInvocationParameter(
                invocation_name="tool_config",
                label="Tool Config",
                canonical_name=CanonicalParameterName.TOOL_CHOICE,
            ),
        ]

    async def chat_completion_create(
        self,
        messages: list[
            tuple[ChatCompletionMessageRole, str, Optional[str], Optional[list[JSONScalarType]]]
        ],
        tools: list[JSONScalarType],
        **invocation_parameters: Any,
    ) -> AsyncIterator[ChatCompletionChunk]:
        from google.genai import types

        contents, system_prompt = self._build_google_messages(messages)

        config_dict = invocation_parameters.copy()

        if system_prompt:
            config_dict["system_instruction"] = system_prompt

        if tools:
            function_declarations = [types.FunctionDeclaration(**tool) for tool in tools]
            config_dict["tools"] = [types.Tool(function_declarations=function_declarations)]

        config = types.GenerateContentConfig.model_validate(config_dict)
        stream = await self.client.aio.models.generate_content_stream(
            model=f"models/{self.model_name}",
            contents=contents,
            config=config,
        )
        async for event in stream:
            self._attributes.update(
                {
                    LLM_TOKEN_COUNT_PROMPT: event.usage_metadata.prompt_token_count,
                    LLM_TOKEN_COUNT_COMPLETION: event.usage_metadata.candidates_token_count,
                    LLM_TOKEN_COUNT_TOTAL: event.usage_metadata.total_token_count,
                }
            )

            if event.candidates:
                candidate = event.candidates[0]
                if candidate.content and candidate.content.parts:
                    for part in candidate.content.parts:
                        if function_call := part.function_call:
                            yield ToolCallChunk(
                                id=function_call.id or "",
                                function=FunctionCallChunk(
                                    name=function_call.name or "",
                                    arguments=json.dumps(function_call.args or {}),
                                ),
                            )
                        elif text := part.text:
                            yield TextChunk(content=text)

    def _build_google_messages(
        self,
        messages: list[tuple[ChatCompletionMessageRole, str, Optional[str], Optional[list[str]]]],
    ) -> tuple[list["ContentType"], str]:
        """Build Google messages following the standard pattern - process ALL messages."""
        google_messages: list["ContentType"] = []
        system_prompts = []
        for role, content, _tool_call_id, _tool_calls in messages:
            if role == ChatCompletionMessageRole.USER:
                google_messages.append({"role": "user", "parts": [{"text": content}]})
            elif role == ChatCompletionMessageRole.AI:
                google_messages.append({"role": "model", "parts": [{"text": content}]})
            elif role == ChatCompletionMessageRole.SYSTEM:
                system_prompts.append(content)
            elif role == ChatCompletionMessageRole.TOOL:
                raise NotImplementedError
            else:
                assert_never(role)

        return google_messages, "\n".join(system_prompts)


@register_llm_client(
    provider_key=GenerativeProviderKey.GOOGLE,
    model_names=[
        PROVIDER_DEFAULT,
        "gemini-2.5-pro",
        "gemini-2.5-flash",
        "gemini-2.5-flash-lite",
        "gemini-2.5-pro-preview-03-25",
    ],
)
class Gemini25GoogleStreamingClient(GoogleStreamingClient):
    @classmethod
    def supported_invocation_parameters(cls) -> list[InvocationParameter]:
        return [
            BoundedFloatInvocationParameter(
                invocation_name="temperature",
                canonical_name=CanonicalParameterName.TEMPERATURE,
                label="Temperature",
                default_value=1.0,
                min_value=0.0,
                max_value=2.0,
            ),
            IntInvocationParameter(
                invocation_name="max_output_tokens",
                canonical_name=CanonicalParameterName.MAX_COMPLETION_TOKENS,
                label="Max Output Tokens",
            ),
            StringListInvocationParameter(
                invocation_name="stop_sequences",
                canonical_name=CanonicalParameterName.STOP_SEQUENCES,
                label="Stop Sequences",
            ),
            BoundedFloatInvocationParameter(
                invocation_name="top_p",
                canonical_name=CanonicalParameterName.TOP_P,
                label="Top P",
                min_value=0.0,
                max_value=1.0,
            ),
            FloatInvocationParameter(
                invocation_name="top_k",
                label="Top K",
            ),
            JSONInvocationParameter(
                invocation_name="tool_config",
                label="Tool Choice",
                canonical_name=CanonicalParameterName.TOOL_CHOICE,
            ),
        ]


@register_llm_client(
    provider_key=GenerativeProviderKey.GOOGLE,
    model_names=[
        "gemini-3-pro-preview",
    ],
)
class Gemini3GoogleStreamingClient(Gemini25GoogleStreamingClient):
    @classmethod
    def supported_invocation_parameters(cls) -> list[InvocationParameter]:
        return [
            StringInvocationParameter(
                invocation_name="thinking_level",
                label="Thinking Level",
                canonical_name=CanonicalParameterName.REASONING_EFFORT,
            ),
            *super().supported_invocation_parameters(),
        ]

    async def chat_completion_create(
        self,
        messages: list[
            tuple[ChatCompletionMessageRole, str, Optional[str], Optional[list[JSONScalarType]]]
        ],
        tools: list[JSONScalarType],
        **invocation_parameters: Any,
    ) -> AsyncIterator[ChatCompletionChunk]:
        # Extract thinking_level and construct thinking_config
        thinking_level = invocation_parameters.pop("thinking_level", None)

        if thinking_level:
            try:
                import google.genai
                from packaging.version import parse as parse_version

                if parse_version(google.genai.__version__) < parse_version("1.50.0"):
                    raise ImportError
            except (ImportError, AttributeError):
                raise BadRequest(
                    "Reasoning capabilities for Gemini models require `google-genai>=1.50.0` "
                    "and Python >= 3.10."
                )

            # NOTE: as of gemini 1.51.0 medium thinking is not supported
            # but will eventually be added in a future version
            # we are purposefully allowing users to select medium knowing
            # it does not work.
            invocation_parameters["thinking_config"] = {
                "include_thoughts": True,
                "thinking_level": thinking_level.upper(),
            }

        async for chunk in super().chat_completion_create(messages, tools, **invocation_parameters):
            yield chunk


def initialize_playground_clients() -> None:
    """
    Ensure that all playground clients are registered at import time.
    """
    pass


LLM_PROVIDER = SpanAttributes.LLM_PROVIDER
LLM_SYSTEM = SpanAttributes.LLM_SYSTEM
LLM_TOKEN_COUNT_PROMPT = SpanAttributes.LLM_TOKEN_COUNT_PROMPT
LLM_TOKEN_COUNT_COMPLETION = SpanAttributes.LLM_TOKEN_COUNT_COMPLETION
LLM_TOKEN_COUNT_TOTAL = SpanAttributes.LLM_TOKEN_COUNT_TOTAL
LLM_TOKEN_COUNT_PROMPT_DETAILS_CACHE_READ = SpanAttributes.LLM_TOKEN_COUNT_PROMPT_DETAILS_CACHE_READ
LLM_TOKEN_COUNT_PROMPT_DETAILS_CACHE_WRITE = (
    SpanAttributes.LLM_TOKEN_COUNT_PROMPT_DETAILS_CACHE_WRITE
)
LLM_TOKEN_COUNT_PROMPT_DETAILS_AUDIO = SpanAttributes.LLM_TOKEN_COUNT_PROMPT_DETAILS_AUDIO
LLM_TOKEN_COUNT_COMPLETION_DETAILS_REASONING = (
    SpanAttributes.LLM_TOKEN_COUNT_COMPLETION_DETAILS_REASONING
)
LLM_TOKEN_COUNT_COMPLETION_DETAILS_AUDIO = SpanAttributes.LLM_TOKEN_COUNT_COMPLETION_DETAILS_AUDIO


class _HttpxClient(wrapt.ObjectProxy):  # type: ignore
    def __init__(self, wrapped: httpx.AsyncClient, attributes: MutableMapping[str, Any]):
        super().__init__(wrapped)
        self._self_attributes = attributes

    async def send(self, request: httpx.Request, **kwargs: Any) -> Any:
        self._self_attributes["url.full"] = str(request.url)
        self._self_attributes["url.path"] = request.url.path.removeprefix(self.base_url.path)
        response = await self.__wrapped__.send(request, **kwargs)
        return response



from sqlalchemy import select
from strawberry.dataloader import DataLoader
from typing_extensions import TypeAlias

from phoenix.db import models
from phoenix.server.types import DbSessionFactory

ExperimentRunID: TypeAlias = int
DatasetExampleID: TypeAlias = int
DatasetVersionID: TypeAlias = int
Key: TypeAlias = ExperimentRunID
Result: TypeAlias = tuple[models.DatasetExample, DatasetVersionID]


class DatasetExamplesAndVersionsByExperimentRunDataLoader(DataLoader[Key, Result]):
    def __init__(self, db: DbSessionFactory) -> None:
        super().__init__(load_fn=self._load_fn)
        self._db = db

    async def _load_fn(self, keys: list[Key]) -> list[Result]:
        experiment_run_ids = set(keys)
        examples_and_versions_query = (
            select(
                models.ExperimentRun.id.label("experiment_run_id"),
                models.DatasetExample,
                models.Experiment.dataset_version_id.label("dataset_version_id"),
            )
            .select_from(models.ExperimentRun)
            .join(
                models.DatasetExample,
                models.DatasetExample.id == models.ExperimentRun.dataset_example_id,
            )
            .join(
                models.Experiment,
                models.Experiment.id == models.ExperimentRun.experiment_id,
            )
            .where(models.ExperimentRun.id.in_(experiment_run_ids))
        )
        async with self._db() as session:
            examples_and_versions = {
                experiment_run_id: (example, version_id)
                for experiment_run_id, example, version_id in (
                    await session.execute(examples_and_versions_query)
                ).all()
            }

        return [examples_and_versions[key] for key in keys]



from collections import defaultdict
from typing import Optional

import numpy as np
from aioitertools.itertools import groupby
from sqlalchemy import select
from strawberry.dataloader import DataLoader
from typing_extensions import TypeAlias

from phoenix.db import models
from phoenix.metrics.retrieval_metrics import RetrievalMetrics
from phoenix.server.api.types.DocumentRetrievalMetrics import DocumentRetrievalMetrics
from phoenix.server.types import DbSessionFactory

RowId: TypeAlias = int
NumDocs: TypeAlias = int
EvalName: TypeAlias = Optional[str]

Key: TypeAlias = tuple[RowId, EvalName, NumDocs]
Result: TypeAlias = list[DocumentRetrievalMetrics]


class DocumentRetrievalMetricsDataLoader(DataLoader[Key, Result]):
    def __init__(self, db: DbSessionFactory) -> None:
        super().__init__(load_fn=self._load_fn)
        self._db = db

    async def _load_fn(self, keys: list[Key]) -> list[Result]:
        mda = models.DocumentAnnotation
        stmt = (
            select(
                mda.span_rowid,
                mda.name,
                mda.score,
                mda.document_position,
            )
            .where(mda.score != None)  # noqa: E711
            .where(mda.annotator_kind == "LLM")
            .where(mda.document_position >= 0)
            .order_by(mda.span_rowid, mda.name)
        )
        # Using CTE with VALUES clause is possible in SQLite, but not in
        # SQLAlchemy v2.0.29, hence the workaround below with over-fetching.
        # We could use CTE with VALUES for postgresql, but for now we'll keep
        # it simple and just use one approach for all backends.
        all_row_ids = {row_id for row_id, _, _ in keys}
        stmt = stmt.where(mda.span_rowid.in_(all_row_ids))
        all_eval_names = {eval_name for _, eval_name, _ in keys}
        if None not in all_eval_names:
            stmt = stmt.where(mda.name.in_(all_eval_names))
        max_position = max(num_docs for _, _, num_docs in keys)
        stmt = stmt.where(mda.document_position < max_position)
        results: dict[Key, Result] = {key: [] for key in keys}
        requested_num_docs: defaultdict[tuple[RowId, EvalName], set[NumDocs]] = defaultdict(set)
        for row_id, eval_name, num_docs in results.keys():
            requested_num_docs[(row_id, eval_name)].add(num_docs)
        async with self._db() as session:
            data = await session.stream(stmt)
            async for (span_rowid, name), group in groupby(data, lambda r: (r.span_rowid, r.name)):
                # We need to fulfill two types of potential requests: 1. when it
                # specifies an evaluation name, and 2. when it doesn't care about
                # the evaluation name by specifying None.
                max_requested_num_docs = max(
                    (
                        num_docs
                        for eval_name in (name, None)
                        for num_docs in (requested_num_docs.get((span_rowid, eval_name)) or ())
                    ),
                    default=0,
                )
                if max_requested_num_docs <= 0:
                    # We have over-fetched. Skip this group.
                    continue
                scores = [np.nan] * max_requested_num_docs
                for row in group:
                    # Length check is necessary due to over-fetching.
                    if row.document_position < len(scores):
                        scores[row.document_position] = row.score
                for eval_name in (name, None):
                    for num_docs in requested_num_docs.get((span_rowid, eval_name)) or ():
                        metrics = RetrievalMetrics(scores[:num_docs])
                        doc_metrics = DocumentRetrievalMetrics(
                            evaluation_name=name, metrics=metrics
                        )
                        key = (span_rowid, eval_name, num_docs)
                        results[key].append(doc_metrics)
        # Make sure to copy the result, so we don't return the same list
        # object to two different requesters.
        return [results[key].copy() for key in keys]



from enum import Enum, auto
from typing import Callable, Mapping, cast

import strawberry

from phoenix.metrics.wrappers import SkEval


@strawberry.enum
class PerformanceMetric(Enum):
    accuracyScore = auto()


PERFORMANCE_METRIC_FUNCTIONS: Mapping[PerformanceMetric, Callable[..., float]] = {
    PerformanceMetric.accuracyScore: cast(Callable[..., float], SkEval.accuracy_score),
}



from collections import defaultdict

from sqlalchemy import select
from strawberry.dataloader import DataLoader
from typing_extensions import TypeAlias

from phoenix.db.models import ProjectSessionAnnotation
from phoenix.server.types import DbSessionFactory

ProjectSessionId: TypeAlias = int
Key: TypeAlias = ProjectSessionId
Result: TypeAlias = list[ProjectSessionAnnotation]


class SessionAnnotationsBySessionDataLoader(DataLoader[Key, Result]):
    def __init__(self, db: DbSessionFactory) -> None:
        super().__init__(load_fn=self._load_fn)
        self._db = db

    async def _load_fn(self, keys: list[Key]) -> list[Result]:
        annotations_by_id: defaultdict[Key, Result] = defaultdict(list)
        async with self._db() as session:
            async for annotation in await session.stream_scalars(
                select(ProjectSessionAnnotation).where(
                    ProjectSessionAnnotation.project_session_id.in_(keys)
                )
            ):
                annotations_by_id[annotation.project_session_id].append(annotation)
        return [annotations_by_id[key] for key in keys]



from typing import Optional

from fastapi import APIRouter, Depends, HTTPException, Query
from pydantic import Field
from sqlalchemy import select
from starlette.requests import Request
from strawberry.relay import GlobalID

from phoenix.db import models
from phoenix.db.helpers import SupportedSQLDialect
from phoenix.db.insertion.helpers import as_kv, insert_on_conflict
from phoenix.server.api.routers.v1.annotations import SpanDocumentAnnotationData
from phoenix.server.api.types.DocumentAnnotation import DocumentAnnotation
from phoenix.server.authorization import is_not_locked
from phoenix.server.bearer_auth import PhoenixUser
from phoenix.server.dml_event import DocumentAnnotationInsertEvent

from .models import V1RoutesBaseModel
from .utils import RequestBody, ResponseBody, add_errors_to_responses

# Since the document annotations are spans related, we place it under spans
router = APIRouter(tags=["spans"])


class AnnotateSpanDocumentsRequestBody(RequestBody[list[SpanDocumentAnnotationData]]):
    pass


class InsertedSpanDocumentAnnotation(V1RoutesBaseModel):
    id: str = Field(description="The ID of the inserted span document annotation")


class AnnotateSpanDocumentsResponseBody(ResponseBody[list[InsertedSpanDocumentAnnotation]]):
    pass


@router.post(
    "/document_annotations",
    dependencies=[Depends(is_not_locked)],
    operation_id="annotateSpanDocuments",
    responses=add_errors_to_responses(
        [
            {
                "status_code": 404,
                "description": "Span not found",
            },
            {
                "status_code": 422,
                "description": "Invalid request - non-empty identifier not supported",
            },
        ]
    ),
    response_description="Span document annotation inserted successfully",
    include_in_schema=True,
)
async def annotate_span_documents(
    request: Request,
    request_body: AnnotateSpanDocumentsRequestBody,
    sync: bool = Query(
        default=False, description="If set to true, the annotations are inserted synchronously."
    ),
) -> AnnotateSpanDocumentsResponseBody:
    if not request_body.data:
        return AnnotateSpanDocumentsResponseBody(data=[])

    # Validate that identifiers are empty or only whitespace
    for annotation in request_body.data:
        if annotation.identifier.strip():
            raise HTTPException(
                detail=f"Non-empty identifier '{annotation.identifier}' is not supported",
                status_code=422,  # Unprocessable Entity
            )

    user_id: Optional[int] = None
    if request.app.state.authentication_enabled and isinstance(request.user, PhoenixUser):
        user_id = int(request.user.identity)

    span_document_annotations = request_body.data

    precursors = [
        annotation.as_precursor(user_id=user_id) for annotation in span_document_annotations
    ]
    if not sync:
        await request.state.enqueue_annotations(*precursors)
        return AnnotateSpanDocumentsResponseBody(data=[])

    span_ids = {p.span_id for p in precursors}
    # Account for the fact that the spans could arrive after the annotation
    async with request.app.state.db() as session:
        existing_spans = {
            span_id: (id_, num_docs)
            async for span_id, id_, num_docs in await session.stream(
                select(models.Span.span_id, models.Span.id, models.Span.num_documents).filter(
                    models.Span.span_id.in_(span_ids)
                )
            )
        }

        missing_span_ids = span_ids - set(existing_spans.keys())
        # We prefer to fail the entire operation if there are missing spans in sync mode
        if missing_span_ids:
            raise HTTPException(
                detail=f"Spans with IDs {', '.join(missing_span_ids)} do not exist.",
                status_code=404,
            )

        # Validate that document positions are within bounds
        for annotation in span_document_annotations:
            _, num_docs = existing_spans[annotation.span_id]
            if annotation.document_position not in range(num_docs):
                raise HTTPException(
                    detail=f"Document position {annotation.document_position} is out of bounds for "
                    f"span {annotation.span_id} (max: {num_docs - 1})",
                    status_code=422,  # Unprocessable Entity
                )

        inserted_document_annotation_ids = []
        dialect = SupportedSQLDialect(session.bind.dialect.name)
        for anno in precursors:
            span_rowid, _ = existing_spans[anno.span_id]
            values = dict(as_kv(anno.as_insertable(span_rowid).row))
            span_document_annotation_id = await session.scalar(
                insert_on_conflict(
                    values,
                    dialect=dialect,
                    table=models.DocumentAnnotation,
                    unique_by=("name", "span_rowid", "identifier", "document_position"),
                    constraint_name="uq_document_annotations_name_span_rowid_document_pos_identifier",
                ).returning(models.DocumentAnnotation.id)
            )
            inserted_document_annotation_ids.append(span_document_annotation_id)

    # We queue an event to let the application know that annotations have changed
    request.state.event_queue.put(
        DocumentAnnotationInsertEvent(tuple(inserted_document_annotation_ids))
    )
    return AnnotateSpanDocumentsResponseBody(
        data=[
            InsertedSpanDocumentAnnotation(id=str(GlobalID(DocumentAnnotation.__name__, str(id_))))
            for id_ in inserted_document_annotation_ids
        ]
    )



from __future__ import annotations

import warnings
from typing import Optional

from fastapi import APIRouter, Depends, HTTPException, Query
from pydantic import Field
from sqlalchemy import select
from starlette.requests import Request

from phoenix.db import models
from phoenix.db.helpers import SupportedSQLDialect
from phoenix.db.insertion.helpers import as_kv, insert_on_conflict
from phoenix.server.api.routers.v1.models import V1RoutesBaseModel
from phoenix.server.authorization import is_not_locked
from phoenix.server.bearer_auth import PhoenixUser

from .annotations import SessionAnnotationData
from .utils import RequestBody, ResponseBody, add_errors_to_responses

router = APIRouter(tags=["sessions"])


class InsertedSessionAnnotation(V1RoutesBaseModel):
    id: str = Field(description="The ID of the inserted session annotation")


class AnnotateSessionsRequestBody(RequestBody[list[SessionAnnotationData]]):
    pass


class AnnotateSessionsResponseBody(ResponseBody[list[InsertedSessionAnnotation]]):
    pass


@router.post(
    "/session_annotations",
    dependencies=[Depends(is_not_locked)],
    operation_id="annotateSessions",
    summary="Create session annotations",
    responses=add_errors_to_responses([{"status_code": 404, "description": "Session not found"}]),
    response_description="Session annotations inserted successfully",
    include_in_schema=True,
)
async def annotate_sessions(
    request: Request,
    request_body: AnnotateSessionsRequestBody,
    sync: bool = Query(default=False, description="If true, fulfill request synchronously."),
) -> AnnotateSessionsResponseBody:
    if not request_body.data:
        return AnnotateSessionsResponseBody(data=[])

    user_id: Optional[int] = None
    if request.app.state.authentication_enabled and isinstance(request.user, PhoenixUser):
        user_id = int(request.user.identity)

    session_annotations = request_body.data
    filtered_session_annotations = list(filter(lambda d: d.name != "note", session_annotations))
    if len(filtered_session_annotations) != len(session_annotations):
        warnings.warn(
            (
                "Session annotations with the name 'note' are not supported in this endpoint. "
                "They will be ignored."
            ),
            UserWarning,
        )
    precursors = [d.as_precursor(user_id=user_id) for d in filtered_session_annotations]
    if not sync:
        await request.state.enqueue_annotations(*precursors)
        return AnnotateSessionsResponseBody(data=[])

    session_ids = {p.session_id for p in precursors}
    async with request.app.state.db() as session:
        existing_sessions = {
            session_id: rowid
            async for session_id, rowid in await session.stream(
                select(models.ProjectSession.session_id, models.ProjectSession.id).filter(
                    models.ProjectSession.session_id.in_(session_ids)
                )
            )
        }

    missing_session_ids = session_ids - set(existing_sessions.keys())
    # We prefer to fail the entire operation if there are missing sessions in sync mode
    if missing_session_ids:
        raise HTTPException(
            detail=f"Sessions with IDs {', '.join(missing_session_ids)} do not exist.",
            status_code=404,
        )

    async with request.app.state.db() as session:
        inserted_ids = []
        dialect = SupportedSQLDialect(session.bind.dialect.name)
        for p in precursors:
            values = dict(as_kv(p.as_insertable(existing_sessions[p.session_id]).row))
            session_annotation_id = await session.scalar(
                insert_on_conflict(
                    values,
                    dialect=dialect,
                    table=models.ProjectSessionAnnotation,
                    unique_by=("name", "project_session_id", "identifier"),
                ).returning(models.ProjectSessionAnnotation.id)
            )
            inserted_ids.append(session_annotation_id)

    return AnnotateSessionsResponseBody(
        data=[InsertedSessionAnnotation(id=str(inserted_id)) for inserted_id in inserted_ids]
    )



import operator
from enum import Enum, auto
from typing import Any, Optional

import strawberry
from sqlalchemy import ColumnElement, Select, and_, func, literal, or_, select, tuple_
from sqlalchemy.sql.selectable import NamedFromClause
from strawberry import Maybe
from typing_extensions import assert_never

from phoenix.db import models
from phoenix.server.api.types.pagination import (
    Cursor,
    CursorSortColumn,
    CursorSortColumnDataType,
    CursorSortColumnValue,
)
from phoenix.server.api.types.SortDir import SortDir


@strawberry.enum
class ExperimentRunMetric(Enum):
    latencyMs = auto()


@strawberry.input(one_of=True)
class ExperimentRunColumn:
    metric: Maybe[ExperimentRunMetric]
    annotation_name: Maybe[str]


@strawberry.input(description="The sort key and direction for experiment run connections")
class ExperimentRunSort:
    col: ExperimentRunColumn
    dir: SortDir


def get_experiment_run_cursor(
    run: models.ExperimentRun, annotation_score: Optional[float], sort: Optional[ExperimentRunSort]
) -> Cursor:
    sort_column: Optional[CursorSortColumn] = None
    if sort:
        if sort.col.metric:
            metric = sort.col.metric.value
            assert metric is not None
            if metric is ExperimentRunMetric.latencyMs:
                sort_column = CursorSortColumn(
                    type=CursorSortColumnDataType.FLOAT,
                    value=run.latency_ms,
                )
            else:
                assert_never(metric)
        elif sort.col.annotation_name:
            data_type = (
                CursorSortColumnDataType.FLOAT
                if annotation_score is not None
                else CursorSortColumnDataType.NULL
            )
            sort_column = CursorSortColumn(
                type=data_type,
                value=annotation_score,
            )
    return Cursor(rowid=run.id, sort_column=sort_column)


def add_order_by_and_page_start_to_query(
    query: Select[Any],
    sort: Optional[ExperimentRunSort],
    experiment_rowid: int,
    after_experiment_run_rowid: Optional[int],
    after_sort_column_value: Optional[CursorSortColumnValue] = None,
) -> Select[Any]:
    mean_annotation_scores: Optional[NamedFromClause] = None
    if sort and sort.col.annotation_name:
        annotation_name = sort.col.annotation_name.value
        assert annotation_name is not None
        mean_annotation_scores = _get_mean_annotation_scores_subquery(annotation_name)
    order_by_columns = _get_order_by_columns(
        sort=sort, experiment_rowid=experiment_rowid, mean_annotation_scores=mean_annotation_scores
    )
    query = query.order_by(*order_by_columns)
    if after_experiment_run_rowid is not None:
        query = _add_after_expression(
            query=query,
            sort=sort,
            experiment_run_rowid=after_experiment_run_rowid,
            after_sort_column_value=after_sort_column_value,
            mean_annotation_scores=mean_annotation_scores,
        )
    query = _add_joins_and_selects_to_query(
        query=query,
        sort=sort,
        mean_annotation_scores=mean_annotation_scores,
    )
    return query


def _get_order_by_columns(
    sort: Optional[ExperimentRunSort],
    experiment_rowid: int,
    mean_annotation_scores: Optional[NamedFromClause],
) -> tuple[ColumnElement[Any], ...]:
    if not sort:
        # Ideally, this would sort the runs by (example_id, repetition_number),
        # but this would require making the cursor more complex or adding an additional query
        # to handle the after cursor.
        return (models.ExperimentRun.id.asc(),)
    sort_direction = sort.dir
    if sort.col.metric:
        metric = sort.col.metric.value
        assert metric is not None
        if metric is ExperimentRunMetric.latencyMs:
            if sort_direction is SortDir.asc:
                return (models.ExperimentRun.latency_ms.asc(), models.ExperimentRun.id.asc())
            else:
                return (models.ExperimentRun.latency_ms.desc(), models.ExperimentRun.id.desc())
        else:
            assert_never(metric)
    elif sort.col.annotation_name:
        annotation_name = sort.col.annotation_name.value
        assert annotation_name is not None
        assert mean_annotation_scores is not None
        if sort_direction is SortDir.asc:
            return (
                mean_annotation_scores.c.score.asc().nulls_last(),
                models.ExperimentRun.id.asc(),
            )
        else:
            return (
                mean_annotation_scores.c.score.desc().nulls_last(),
                models.ExperimentRun.id.desc(),
            )
    raise NotImplementedError


def _add_after_expression(
    query: Select[Any],
    sort: Optional[ExperimentRunSort],
    experiment_run_rowid: int,
    after_sort_column_value: Optional[CursorSortColumnValue],
    mean_annotation_scores: Optional[NamedFromClause],
) -> Select[Any]:
    if not sort:
        # Ideally, this would return the runs sorted by (example_id, repetition_number),
        # but this would require making the cursor more complex or adding an additional query.
        return query.where(models.ExperimentRun.id > literal(experiment_run_rowid))
    sort_direction = sort.dir
    compare_fn = operator.gt if sort_direction is SortDir.asc else operator.lt
    if sort.col.metric:
        metric = sort.col.metric.value
        assert metric is not None
        if metric is ExperimentRunMetric.latencyMs:
            assert after_sort_column_value is not None
            return query.where(
                compare_fn(
                    tuple_(models.ExperimentRun.latency_ms, models.ExperimentRun.id),
                    tuple_(
                        literal(after_sort_column_value),
                        literal(experiment_run_rowid),
                    ),
                )
            )
        else:
            assert_never(metric)
    elif sort.col.annotation_name:
        annotation_name = sort.col.annotation_name.value
        assert annotation_name is not None
        assert mean_annotation_scores is not None
        if after_sort_column_value is None:
            return query.where(
                and_(
                    compare_fn(models.ExperimentRun.id, literal(experiment_run_rowid)),
                    mean_annotation_scores.c.score.is_(None),
                )
            )
        else:
            return query.where(
                or_(
                    compare_fn(
                        tuple_(mean_annotation_scores.c.score, models.ExperimentRun.id),
                        tuple_(
                            literal(after_sort_column_value),
                            literal(experiment_run_rowid),
                        ),
                    ),
                    mean_annotation_scores.c.score.is_(None),
                )
            )
    raise NotImplementedError


def _get_mean_annotation_scores_subquery(annotation_name: str) -> NamedFromClause:
    return (
        select(
            func.avg(models.ExperimentRunAnnotation.score).label("score"),
            models.ExperimentRunAnnotation.experiment_run_id.label("experiment_run_id"),
        )
        .select_from(models.ExperimentRunAnnotation)
        .join(
            models.ExperimentRun,
            models.ExperimentRunAnnotation.experiment_run_id == models.ExperimentRun.id,
        )
        .where(models.ExperimentRunAnnotation.name == annotation_name)
        .group_by(models.ExperimentRunAnnotation.experiment_run_id)
        .subquery()
        .alias("mean_annotation_scores")
    )


def _add_joins_and_selects_to_query(
    query: Select[tuple[models.ExperimentRun]],
    sort: Optional[ExperimentRunSort],
    mean_annotation_scores: Optional[NamedFromClause],
) -> Select[tuple[models.ExperimentRun]]:
    if not sort:
        return query
    if sort.col.metric:
        metric = sort.col.metric.value
        assert metric is not None
        if metric is ExperimentRunMetric.latencyMs:
            return query
        else:
            assert_never(metric)
    elif sort.col.annotation_name:
        annotation_name = sort.col.annotation_name.value
        assert annotation_name is not None
        assert mean_annotation_scores is not None
        query = query.join(
            mean_annotation_scores,
            mean_annotation_scores.c.experiment_run_id == models.ExperimentRun.id,
            isouter=True,
        )
        query = query.add_columns(
            mean_annotation_scores.c.score.label("score")
        )  # the score must be in the select so that the value can be included in the cursor
        return query
    raise NotImplementedError



from typing import Annotated

from pydantic import Field, RootModel


class Identifier(RootModel[str]):
    root: Annotated[str, Field(pattern=r"^[a-z0-9]([_a-z0-9-]*[a-z0-9])?$")]

    def __hash__(self) -> int:
        return hash(self.root)



from typing import Optional

from sqlalchemy import func, select
from strawberry.dataloader import DataLoader
from typing_extensions import TypeAlias

from phoenix.db import models
from phoenix.server.types import DbSessionFactory

ExperimentID: TypeAlias = int
ErrorRate: TypeAlias = float
Key: TypeAlias = ExperimentID
Result: TypeAlias = Optional[ErrorRate]


class ExperimentErrorRatesDataLoader(DataLoader[Key, Result]):
    def __init__(
        self,
        db: DbSessionFactory,
    ) -> None:
        super().__init__(load_fn=self._load_fn)
        self._db = db

    async def _load_fn(self, keys: list[Key]) -> list[Result]:
        experiment_ids = keys
        average_repetition_error_rates_subquery = (
            select(
                models.ExperimentRun.experiment_id.label("experiment_id"),
                (
                    func.count(models.ExperimentRun.error) / func.count(models.ExperimentRun.id)
                ).label("average_repetition_error_rate"),
            )
            .where(models.ExperimentRun.experiment_id.in_(experiment_ids))
            .group_by(models.ExperimentRun.dataset_example_id, models.ExperimentRun.experiment_id)
            .subquery()
            .alias("average_repetition_error_rates")
        )
        average_run_error_rates_query = select(
            average_repetition_error_rates_subquery.c.experiment_id,
            func.avg(average_repetition_error_rates_subquery.c.average_repetition_error_rate).label(
                "average_run_error_rates"
            ),
        ).group_by(average_repetition_error_rates_subquery.c.experiment_id)
        async with self._db() as session:
            average_run_error_rates = {
                experiment_id: error_rate
                async for experiment_id, error_rate in await session.stream(
                    average_run_error_rates_query
                )
            }
        return [average_run_error_rates.get(experiment_id) for experiment_id in experiment_ids]



from typing import Optional

import strawberry

from phoenix.server.api.types.GenerativeProvider import GenerativeProviderKey


@strawberry.interface
class ModelInterface:
    @strawberry.field
    async def name(self) -> str:
        raise NotImplementedError

    @strawberry.field
    async def provider_key(self) -> Optional[GenerativeProviderKey]:
        raise NotImplementedError



from typing import Optional

from sqlalchemy import func, select
from strawberry.dataloader import DataLoader
from typing_extensions import TypeAlias

from phoenix.db import models
from phoenix.server.types import DbSessionFactory

ExperimentID: TypeAlias = int
RunLatency: TypeAlias = Optional[float]
Key: TypeAlias = ExperimentID
Result: TypeAlias = RunLatency


class AverageExperimentRunLatencyDataLoader(DataLoader[Key, Result]):
    def __init__(
        self,
        db: DbSessionFactory,
    ) -> None:
        super().__init__(load_fn=self._load_fn)
        self._db = db

    async def _load_fn(self, keys: list[Key]) -> list[Result]:
        experiment_ids = keys
        average_repetition_latency_ms = (
            select(
                models.ExperimentRun.experiment_id.label("experiment_id"),
                func.avg(models.ExperimentRun.latency_ms).label("average_repetition_latency_ms"),
            )
            .select_from(models.ExperimentRun)
            .where(models.ExperimentRun.experiment_id.in_(experiment_ids))
            .group_by(models.ExperimentRun.dataset_example_id, models.ExperimentRun.experiment_id)
            .subquery()
        )
        query = select(
            average_repetition_latency_ms.c.experiment_id,
            func.avg(average_repetition_latency_ms.c.average_repetition_latency_ms).label(
                "average_run_latency_ms"
            ),
        ).group_by(average_repetition_latency_ms.c.experiment_id)
        async with self._db() as session:
            average_run_latencies_ms = {
                experiment_id: average_run_latency_ms
                async for experiment_id, average_run_latency_ms in await session.stream(query)
            }
        return [average_run_latencies_ms.get(experiment_id) for experiment_id in keys]



import re
from typing import Optional

import strawberry
from strawberry import UNSET

from phoenix.server.api.exceptions import BadRequest


@strawberry.input
class CreateProjectInput:
    name: str
    description: Optional[str] = UNSET
    gradient_start_color: Optional[str] = UNSET
    gradient_end_color: Optional[str] = UNSET

    def __post_init__(self) -> None:
        if not self.name.strip():
            raise BadRequest("Name cannot be empty")
        if self.gradient_start_color and not re.match(
            r"^#([0-9a-fA-F]{6})$", self.gradient_start_color
        ):
            raise BadRequest("Gradient start color must be a valid hex color")
        if self.gradient_end_color and not re.match(
            r"^#([0-9a-fA-F]{6})$", self.gradient_end_color
        ):
            raise BadRequest("Gradient end color must be a valid hex color")



from enum import Enum

import strawberry


@strawberry.enum
class ProjectFilterColumn(Enum):
    name = "name"


@strawberry.input(description="The filter key and value for project connections")
class ProjectFilter:
    col: ProjectFilterColumn
    value: str



from typing import Optional, Union, cast

import pandas as pd
import strawberry
from strawberry import Private

from phoenix.db import models
from phoenix.server.api.types.LabelFraction import LabelFraction

AnnotationType = Union[models.SpanAnnotation, models.TraceAnnotation]


@strawberry.type
class AnnotationSummary:
    name: str
    df: Private[pd.DataFrame]
    simple_avg: Private[bool] = False

    @strawberry.field
    def count(self) -> int:
        return cast(int, self.df.record_count.sum())

    @strawberry.field
    def labels(self) -> list[str]:
        unique_labels = self.df["label"].dropna().unique()
        return [str(label) for label in unique_labels]

    @strawberry.field
    def label_fractions(self) -> list[LabelFraction]:
        if self.simple_avg:
            if not (n := self.df.label_count.sum()):
                return []
            return [
                LabelFraction(
                    label=cast(str, row.label),
                    fraction=row.label_count / n,
                )
                for row in self.df.loc[
                    self.df.label.notna(),
                    ["label", "label_count"],
                ].itertuples()
            ]
        return [
            LabelFraction(
                label=row.label,
                fraction=float(row.avg_label_fraction),
            )
            for row in self.df.itertuples()
            if row.label is not None
        ]

    @strawberry.field
    def mean_score(self) -> Optional[float]:
        if self.simple_avg:
            if not (n := self.df.score_count.sum()):
                return None
            return cast(float, self.df.score_sum.sum() / n)
        avg_scores = self.df["avg_score"].dropna()
        if avg_scores.empty:
            return None
        return float(avg_scores.mean())  # all avg_scores should be the same

    @strawberry.field
    def score_count(self) -> int:
        return cast(int, self.df.score_count.sum())

    @strawberry.field
    def label_count(self) -> int:
        return cast(int, self.df.label_count.sum())



from collections.abc import AsyncIterable
from datetime import datetime
from typing import Optional, cast

import strawberry
from sqlalchemy import Text, and_, func, or_, select
from sqlalchemy.sql.functions import count
from strawberry import UNSET
from strawberry.relay import Connection, GlobalID, Node, NodeID
from strawberry.scalars import JSON
from strawberry.types import Info

from phoenix.db import models
from phoenix.server.api.context import Context
from phoenix.server.api.exceptions import BadRequest
from phoenix.server.api.input_types.DatasetVersionSort import DatasetVersionSort
from phoenix.server.api.types.DatasetExample import DatasetExample
from phoenix.server.api.types.DatasetExperimentAnnotationSummary import (
    DatasetExperimentAnnotationSummary,
)
from phoenix.server.api.types.DatasetLabel import DatasetLabel
from phoenix.server.api.types.DatasetSplit import DatasetSplit
from phoenix.server.api.types.DatasetVersion import DatasetVersion
from phoenix.server.api.types.Experiment import Experiment, to_gql_experiment
from phoenix.server.api.types.node import from_global_id_with_expected_type
from phoenix.server.api.types.pagination import (
    ConnectionArgs,
    CursorString,
    connection_from_list,
)
from phoenix.server.api.types.SortDir import SortDir


@strawberry.type
class Dataset(Node):
    id: NodeID[int]
    db_record: strawberry.Private[Optional[models.Dataset]] = None

    def __post_init__(self) -> None:
        if self.db_record and self.id != self.db_record.id:
            raise ValueError("Dataset ID mismatch")

    @strawberry.field
    async def name(
        self,
        info: Info[Context, None],
    ) -> str:
        if self.db_record:
            val = self.db_record.name
        else:
            val = await info.context.data_loaders.dataset_fields.load(
                (self.id, models.Dataset.name),
            )
        return val

    @strawberry.field
    async def description(
        self,
        info: Info[Context, None],
    ) -> Optional[str]:
        if self.db_record:
            val = self.db_record.description
        else:
            val = await info.context.data_loaders.dataset_fields.load(
                (self.id, models.Dataset.description),
            )
        return val

    @strawberry.field
    async def metadata(
        self,
        info: Info[Context, None],
    ) -> JSON:
        if self.db_record:
            val = self.db_record.metadata_
        else:
            val = await info.context.data_loaders.dataset_fields.load(
                (self.id, models.Dataset.metadata_),
            )
        return val

    @strawberry.field
    async def created_at(
        self,
        info: Info[Context, None],
    ) -> datetime:
        if self.db_record:
            val = self.db_record.created_at
        else:
            val = await info.context.data_loaders.dataset_fields.load(
                (self.id, models.Dataset.created_at),
            )
        return val

    @strawberry.field
    async def updated_at(
        self,
        info: Info[Context, None],
    ) -> datetime:
        if self.db_record:
            val = self.db_record.updated_at
        else:
            val = await info.context.data_loaders.dataset_fields.load(
                (self.id, models.Dataset.updated_at),
            )
        return val

    @strawberry.field
    async def versions(
        self,
        info: Info[Context, None],
        first: Optional[int] = 50,
        last: Optional[int] = UNSET,
        after: Optional[CursorString] = UNSET,
        before: Optional[CursorString] = UNSET,
        sort: Optional[DatasetVersionSort] = UNSET,
    ) -> Connection[DatasetVersion]:
        args = ConnectionArgs(
            first=first,
            after=after if isinstance(after, CursorString) else None,
            last=last,
            before=before if isinstance(before, CursorString) else None,
        )
        async with info.context.db() as session:
            stmt = select(models.DatasetVersion).filter_by(dataset_id=self.id)
            if sort:
                # For now assume the the column names match 1:1 with the enum values
                sort_col = getattr(models.DatasetVersion, sort.col.value)
                if sort.dir is SortDir.desc:
                    stmt = stmt.order_by(sort_col.desc(), models.DatasetVersion.id.desc())
                else:
                    stmt = stmt.order_by(sort_col.asc(), models.DatasetVersion.id.asc())
            else:
                stmt = stmt.order_by(models.DatasetVersion.created_at.desc())
            versions = await session.scalars(stmt)
        data = [DatasetVersion(id=version.id, db_record=version) for version in versions]
        return connection_from_list(data=data, args=args)

    @strawberry.field(
        description="Number of examples in a specific version if version is specified, or in the "
        "latest version if version is not specified."
    )  # type: ignore
    async def example_count(
        self,
        info: Info[Context, None],
        dataset_version_id: Optional[GlobalID] = UNSET,
        split_ids: Optional[list[GlobalID]] = UNSET,
    ) -> int:
        dataset_id = self.id
        version_id = (
            from_global_id_with_expected_type(
                global_id=dataset_version_id,
                expected_type_name=DatasetVersion.__name__,
            )
            if dataset_version_id
            else None
        )

        # Parse split IDs if provided
        split_rowids: Optional[list[int]] = None
        if split_ids:
            split_rowids = []
            for split_id in split_ids:
                try:
                    split_rowid = from_global_id_with_expected_type(
                        global_id=split_id, expected_type_name=models.DatasetSplit.__name__
                    )
                    split_rowids.append(split_rowid)
                except Exception:
                    raise BadRequest(f"Invalid split ID: {split_id}")

        revision_ids = (
            select(func.max(models.DatasetExampleRevision.id))
            .join(models.DatasetExample)
            .where(models.DatasetExample.dataset_id == dataset_id)
            .group_by(models.DatasetExampleRevision.dataset_example_id)
        )
        if version_id:
            version_id_subquery = (
                select(models.DatasetVersion.id)
                .where(models.DatasetVersion.dataset_id == dataset_id)
                .where(models.DatasetVersion.id == version_id)
                .scalar_subquery()
            )
            revision_ids = revision_ids.where(
                models.DatasetExampleRevision.dataset_version_id <= version_id_subquery
            )

        # Build the count query
        if split_rowids:
            # When filtering by splits, count distinct examples that belong to those splits
            stmt = (
                select(count(models.DatasetExample.id.distinct()))
                .join(
                    models.DatasetExampleRevision,
                    onclause=(
                        models.DatasetExample.id == models.DatasetExampleRevision.dataset_example_id
                    ),
                )
                .join(
                    models.DatasetSplitDatasetExample,
                    onclause=(
                        models.DatasetExample.id
                        == models.DatasetSplitDatasetExample.dataset_example_id
                    ),
                )
                .where(models.DatasetExampleRevision.id.in_(revision_ids))
                .where(models.DatasetExampleRevision.revision_kind != "DELETE")
                .where(models.DatasetSplitDatasetExample.dataset_split_id.in_(split_rowids))
            )
        else:
            stmt = (
                select(count(models.DatasetExampleRevision.id))
                .where(models.DatasetExampleRevision.id.in_(revision_ids))
                .where(models.DatasetExampleRevision.revision_kind != "DELETE")
            )

        async with info.context.db() as session:
            return (await session.scalar(stmt)) or 0

    @strawberry.field
    async def examples(
        self,
        info: Info[Context, None],
        dataset_version_id: Optional[GlobalID] = UNSET,
        split_ids: Optional[list[GlobalID]] = UNSET,
        first: Optional[int] = 50,
        last: Optional[int] = UNSET,
        after: Optional[CursorString] = UNSET,
        before: Optional[CursorString] = UNSET,
        filter: Optional[str] = UNSET,
    ) -> Connection[DatasetExample]:
        args = ConnectionArgs(
            first=first,
            after=after if isinstance(after, CursorString) else None,
            last=last,
            before=before if isinstance(before, CursorString) else None,
        )
        dataset_id = self.id
        version_id = (
            from_global_id_with_expected_type(
                global_id=dataset_version_id, expected_type_name=DatasetVersion.__name__
            )
            if dataset_version_id
            else None
        )

        # Parse split IDs if provided
        split_rowids: Optional[list[int]] = None
        if split_ids:
            split_rowids = []
            for split_id in split_ids:
                try:
                    split_rowid = from_global_id_with_expected_type(
                        global_id=split_id, expected_type_name=models.DatasetSplit.__name__
                    )
                    split_rowids.append(split_rowid)
                except Exception:
                    raise BadRequest(f"Invalid split ID: {split_id}")

        revision_ids = (
            select(func.max(models.DatasetExampleRevision.id))
            .join(models.DatasetExample)
            .where(models.DatasetExample.dataset_id == dataset_id)
            .group_by(models.DatasetExampleRevision.dataset_example_id)
        )
        if version_id:
            version_id_subquery = (
                select(models.DatasetVersion.id)
                .where(models.DatasetVersion.dataset_id == dataset_id)
                .where(models.DatasetVersion.id == version_id)
                .scalar_subquery()
            )
            revision_ids = revision_ids.where(
                models.DatasetExampleRevision.dataset_version_id <= version_id_subquery
            )
        query = (
            select(models.DatasetExample)
            .join(
                models.DatasetExampleRevision,
                onclause=models.DatasetExample.id
                == models.DatasetExampleRevision.dataset_example_id,
            )
            .where(
                and_(
                    models.DatasetExampleRevision.id.in_(revision_ids),
                    models.DatasetExampleRevision.revision_kind != "DELETE",
                )
            )
            .order_by(models.DatasetExample.id.desc())
        )

        # Filter by split IDs if provided
        if split_rowids:
            query = (
                query.join(
                    models.DatasetSplitDatasetExample,
                    onclause=(
                        models.DatasetExample.id
                        == models.DatasetSplitDatasetExample.dataset_example_id
                    ),
                )
                .where(models.DatasetSplitDatasetExample.dataset_split_id.in_(split_rowids))
                .distinct()
            )
        # Apply filter if provided - search through JSON fields (input, output, metadata)
        if filter is not UNSET and filter:
            # Create a filter that searches for the filter string in JSON fields
            # Using PostgreSQL's JSON operators for case-insensitive text search
            filter_condition = or_(
                func.cast(models.DatasetExampleRevision.input, Text).ilike(f"%{filter}%"),
                func.cast(models.DatasetExampleRevision.output, Text).ilike(f"%{filter}%"),
                func.cast(models.DatasetExampleRevision.metadata_, Text).ilike(f"%{filter}%"),
            )
            query = query.where(filter_condition)

        async with info.context.db() as session:
            dataset_examples = [
                DatasetExample(
                    id=example.id,
                    db_record=example,
                    version_id=version_id,
                )
                async for example in await session.stream_scalars(query)
            ]
        return connection_from_list(data=dataset_examples, args=args)

    @strawberry.field
    async def splits(self, info: Info[Context, None]) -> list[DatasetSplit]:
        return [
            DatasetSplit(id=split.id, db_record=split)
            for split in await info.context.data_loaders.dataset_dataset_splits.load(self.id)
        ]

    @strawberry.field(
        description="Number of experiments for a specific version if version is specified, "
        "or for all versions if version is not specified."
    )  # type: ignore
    async def experiment_count(
        self,
        info: Info[Context, None],
        dataset_version_id: Optional[GlobalID] = UNSET,
    ) -> int:
        stmt = select(count(models.Experiment.id)).where(models.Experiment.dataset_id == self.id)
        version_id = (
            from_global_id_with_expected_type(
                global_id=dataset_version_id,
                expected_type_name=DatasetVersion.__name__,
            )
            if dataset_version_id
            else None
        )
        if version_id is not None:
            stmt = stmt.where(models.Experiment.dataset_version_id == version_id)
        async with info.context.db() as session:
            return (await session.scalar(stmt)) or 0

    @strawberry.field
    async def experiments(
        self,
        info: Info[Context, None],
        first: Optional[int] = 50,
        last: Optional[int] = UNSET,
        after: Optional[CursorString] = UNSET,
        before: Optional[CursorString] = UNSET,
        filter_condition: Optional[str] = UNSET,
        filter_ids: Optional[
            list[GlobalID]
        ] = UNSET,  # this is a stopgap until a query DSL is implemented
    ) -> Connection[Experiment]:
        args = ConnectionArgs(
            first=first,
            after=after if isinstance(after, CursorString) else None,
            last=last,
            before=before if isinstance(before, CursorString) else None,
        )
        dataset_id = self.id
        row_number = func.row_number().over(order_by=models.Experiment.id).label("row_number")
        query = (
            select(models.Experiment, row_number)
            .where(models.Experiment.dataset_id == dataset_id)
            .order_by(models.Experiment.id.desc())
        )
        if filter_condition is not UNSET and filter_condition:
            # Search both name and description columns with case-insensitive partial matching
            search_filter = or_(
                models.Experiment.name.ilike(f"%{filter_condition}%"),
                models.Experiment.description.ilike(f"%{filter_condition}%"),
            )
            query = query.where(search_filter)

        if filter_ids:
            filter_rowids = []
            for filter_id in filter_ids:
                try:
                    filter_rowids.append(
                        from_global_id_with_expected_type(
                            global_id=filter_id,
                            expected_type_name=Experiment.__name__,
                        )
                    )
                except ValueError:
                    raise BadRequest(f"Invalid filter ID: {filter_id}")
            query = query.where(models.Experiment.id.in_(filter_rowids))

        async with info.context.db() as session:
            experiments = [
                to_gql_experiment(experiment, sequence_number)
                async for experiment, sequence_number in cast(
                    AsyncIterable[tuple[models.Experiment, int]],
                    await session.stream(query),
                )
            ]
        return connection_from_list(data=experiments, args=args)

    @strawberry.field
    async def experiment_annotation_summaries(
        self, info: Info[Context, None]
    ) -> list[DatasetExperimentAnnotationSummary]:
        dataset_id = self.id
        query = (
            select(
                models.ExperimentRunAnnotation.name.label("annotation_name"),
                func.min(models.ExperimentRunAnnotation.score).label("min_score"),
                func.max(models.ExperimentRunAnnotation.score).label("max_score"),
            )
            .select_from(models.ExperimentRunAnnotation)
            .join(
                models.ExperimentRun,
                models.ExperimentRunAnnotation.experiment_run_id == models.ExperimentRun.id,
            )
            .join(
                models.Experiment,
                models.ExperimentRun.experiment_id == models.Experiment.id,
            )
            .where(models.Experiment.dataset_id == dataset_id)
            .group_by(models.ExperimentRunAnnotation.name)
            .order_by(models.ExperimentRunAnnotation.name)
        )
        async with info.context.db() as session:
            return [
                DatasetExperimentAnnotationSummary(
                    annotation_name=scores_tuple.annotation_name,
                    min_score=scores_tuple.min_score,
                    max_score=scores_tuple.max_score,
                )
                async for scores_tuple in await session.stream(query)
            ]

    @strawberry.field
    async def labels(self, info: Info[Context, None]) -> list[DatasetLabel]:
        return [
            DatasetLabel(id=label.id, db_record=label)
            for label in await info.context.data_loaders.dataset_labels.load(self.id)
        ]

    @strawberry.field
    def last_updated_at(self, info: Info[Context, None]) -> Optional[datetime]:
        return info.context.last_updated_at.get(models.Dataset, self.id)



import gzip
import zlib
from typing import Optional

from fastapi import APIRouter, BackgroundTasks, Depends, Header, HTTPException, Path, Query
from google.protobuf.message import DecodeError
from opentelemetry.proto.collector.trace.v1.trace_service_pb2 import (
    ExportTraceServiceRequest,
    ExportTraceServiceResponse,
)
from pydantic import Field
from sqlalchemy import delete, select
from starlette.concurrency import run_in_threadpool
from starlette.datastructures import State
from starlette.requests import Request
from starlette.responses import Response
from strawberry.relay import GlobalID

from phoenix.db import models
from phoenix.db.helpers import SupportedSQLDialect
from phoenix.db.insertion.helpers import as_kv, insert_on_conflict
from phoenix.server.api.routers.v1.annotations import TraceAnnotationData
from phoenix.server.api.types.node import from_global_id_with_expected_type
from phoenix.server.authorization import is_not_locked
from phoenix.server.bearer_auth import PhoenixUser
from phoenix.server.dml_event import SpanDeleteEvent, TraceAnnotationInsertEvent
from phoenix.server.prometheus import SPAN_QUEUE_REJECTIONS
from phoenix.trace.otel import decode_otlp_span
from phoenix.utilities.project import get_project_name

from .models import V1RoutesBaseModel
from .utils import (
    RequestBody,
    ResponseBody,
    add_errors_to_responses,
)

router = APIRouter(tags=["traces"])


def is_not_at_capacity(request: Request) -> None:
    if request.app.state.span_queue_is_full():
        SPAN_QUEUE_REJECTIONS.inc()
        raise HTTPException(
            detail="Server is at capacity and cannot process more requests",
            status_code=503,
        )


@router.post(
    "/traces",
    dependencies=[Depends(is_not_locked), Depends(is_not_at_capacity)],
    operation_id="addTraces",
    summary="Send traces",
    responses=add_errors_to_responses(
        [
            {
                "status_code": 415,
                "description": (
                    "Unsupported content type (only `application/x-protobuf` is supported)"
                ),
            },
            {"status_code": 422, "description": "Invalid request body"},
            {
                "status_code": 503,
                "description": "Server is at capacity and cannot process more requests",
            },
        ]
    ),
    openapi_extra={
        "requestBody": {
            "required": True,
            "content": {
                "application/x-protobuf": {"schema": {"type": "string", "format": "binary"}}
            },
        }
    },
    include_in_schema=False,
)
async def post_traces(
    request: Request,
    background_tasks: BackgroundTasks,
    content_type: Optional[str] = Header(default=None),
    content_encoding: Optional[str] = Header(default=None),
) -> Response:
    if content_type != "application/x-protobuf":
        raise HTTPException(
            detail=f"Unsupported content type: {content_type}",
            status_code=415,
        )
    if content_encoding and content_encoding not in ("gzip", "deflate"):
        raise HTTPException(
            detail=f"Unsupported content encoding: {content_encoding}",
            status_code=415,
        )
    body = await request.body()
    if content_encoding == "gzip":
        body = await run_in_threadpool(gzip.decompress, body)
    elif content_encoding == "deflate":
        body = await run_in_threadpool(zlib.decompress, body)
    req = ExportTraceServiceRequest()
    try:
        await run_in_threadpool(req.ParseFromString, body)
    except DecodeError:
        raise HTTPException(
            detail="Request body is invalid ExportTraceServiceRequest",
            status_code=422,
        )
    background_tasks.add_task(_add_spans, req, request.state)

    # "The server MUST use the same Content-Type in the response as it received in the request"
    response_message = ExportTraceServiceResponse()
    response_bytes = response_message.SerializeToString()
    return Response(
        content=response_bytes,
        media_type="application/x-protobuf",
        status_code=200,
    )


class AnnotateTracesRequestBody(RequestBody[list[TraceAnnotationData]]):
    data: list[TraceAnnotationData] = Field(description="The trace annotations to be upserted")


class InsertedTraceAnnotation(V1RoutesBaseModel):
    id: str = Field(description="The ID of the inserted trace annotation")


class AnnotateTracesResponseBody(ResponseBody[list[InsertedTraceAnnotation]]):
    pass


@router.post(
    "/trace_annotations",
    dependencies=[Depends(is_not_locked)],
    operation_id="annotateTraces",
    summary="Create trace annotations",
    responses=add_errors_to_responses([{"status_code": 404, "description": "Trace not found"}]),
)
async def annotate_traces(
    request: Request,
    request_body: AnnotateTracesRequestBody,
    sync: bool = Query(default=False, description="If true, fulfill request synchronously."),
) -> AnnotateTracesResponseBody:
    if not request_body.data:
        return AnnotateTracesResponseBody(data=[])

    user_id: Optional[int] = None
    if request.app.state.authentication_enabled and isinstance(request.user, PhoenixUser):
        user_id = int(request.user.identity)

    precursors = [d.as_precursor(user_id=user_id) for d in request_body.data]
    if not sync:
        await request.state.enqueue_annotations(*precursors)
        return AnnotateTracesResponseBody(data=[])

    trace_ids = {p.trace_id for p in precursors}
    async with request.app.state.db() as session:
        existing_traces = {
            trace_id: id_
            async for trace_id, id_ in await session.stream(
                select(models.Trace.trace_id, models.Trace.id).filter(
                    models.Trace.trace_id.in_(trace_ids)
                )
            )
        }

        missing_trace_ids = trace_ids - set(existing_traces.keys())
        if missing_trace_ids:
            raise HTTPException(
                detail=f"Traces with IDs {', '.join(missing_trace_ids)} do not exist.",
                status_code=404,
            )
        inserted_ids = []
        dialect = SupportedSQLDialect(session.bind.dialect.name)
        for p in precursors:
            values = dict(as_kv(p.as_insertable(existing_traces[p.trace_id]).row))
            trace_annotation_id = await session.scalar(
                insert_on_conflict(
                    values,
                    dialect=dialect,
                    table=models.TraceAnnotation,
                    unique_by=("name", "trace_rowid", "identifier"),
                ).returning(models.TraceAnnotation.id)
            )
            inserted_ids.append(trace_annotation_id)
    request.state.event_queue.put(TraceAnnotationInsertEvent(tuple(inserted_ids)))
    return AnnotateTracesResponseBody(
        data=[
            InsertedTraceAnnotation(id=str(GlobalID("TraceAnnotation", str(id_))))
            for id_ in inserted_ids
        ]
    )


async def _add_spans(req: ExportTraceServiceRequest, state: State) -> None:
    for resource_spans in req.resource_spans:
        project_name = get_project_name(resource_spans.resource.attributes)
        for scope_span in resource_spans.scope_spans:
            for otlp_span in scope_span.spans:
                span = await run_in_threadpool(decode_otlp_span, otlp_span)
                await state.enqueue_span(span, project_name)


@router.delete(
    "/traces/{trace_identifier}",
    operation_id="deleteTrace",
    summary="Delete a trace by identifier",
    description=(
        "Delete an entire trace by its identifier. The identifier can be either:\n"
        "1. A Relay node ID (base64-encoded)\n"
        "2. An OpenTelemetry trace_id (hex string)\n\n"
        "This will permanently remove all spans in the trace and their associated data."
    ),
    responses=add_errors_to_responses([404]),
    status_code=204,  # No Content for successful deletion
)
async def delete_trace(
    request: Request,
    trace_identifier: str = Path(
        description="The trace identifier: either a relay GlobalID or OpenTelemetry trace_id"
    ),
) -> None:
    """
    Delete a trace by identifier (relay GlobalID or OpenTelemetry trace_id).

    This endpoint will:
    1. Delete the trace by identifier (relay GlobalID or OpenTelemetry trace_id)
    2. Get project_id from the deletion for cache invalidation
    3. Trigger cache invalidation events
    4. Return 204 No Content on success

    Note: This deletes the entire trace, including all spans, which maintains data consistency
    and avoids orphaned spans or inconsistent cached cumulative fields.
    """
    async with request.app.state.db() as session:
        # Try to parse as GlobalID first, then fall back to trace_id
        try:
            trace_rowid = from_global_id_with_expected_type(
                GlobalID.from_id(trace_identifier),
                "Trace",
            )
            # Delete by database rowid
            delete_stmt = (
                delete(models.Trace)
                .where(models.Trace.id == trace_rowid)
                .returning(models.Trace.project_rowid)
            )
            error_detail = f"Trace with relay ID '{trace_identifier}' not found"
        except Exception:
            # Delete by OpenTelemetry trace_id
            delete_stmt = (
                delete(models.Trace)
                .where(models.Trace.trace_id == trace_identifier)
                .returning(models.Trace.project_rowid)
            )
            error_detail = f"Trace with trace_id '{trace_identifier}' not found"

        project_id = await session.scalar(delete_stmt)

        if project_id is None:
            raise HTTPException(
                status_code=404,
                detail=error_detail,
            )

    # Trigger cache invalidation event
    request.state.event_queue.put(SpanDeleteEvent((project_id,)))

    # Return 204 No Content (successful deletion with no response body)
    return None



from datetime import datetime
from typing import Optional

import strawberry
from strawberry.relay import Node, NodeID
from strawberry.types import Info

from phoenix.db.models import ApiKey as OrmApiKey
from phoenix.server.api.context import Context

from .ApiKey import ApiKey


@strawberry.type
class SystemApiKey(ApiKey, Node):
    id: NodeID[int]
    db_record: strawberry.Private[Optional[OrmApiKey]] = None

    def __post_init__(self) -> None:
        if self.db_record and self.id != self.db_record.id:
            raise ValueError("SystemApiKey ID mismatch")

    @strawberry.field(description="Name of the API key.")  # type: ignore
    async def name(
        self,
        info: Info[Context, None],
    ) -> str:
        if self.db_record:
            val = self.db_record.name
        else:
            val = await info.context.data_loaders.user_api_key_fields.load(
                (self.id, OrmApiKey.name),
            )
        return val

    @strawberry.field(description="Description of the API key.")  # type: ignore
    async def description(
        self,
        info: Info[Context, None],
    ) -> Optional[str]:
        if self.db_record:
            val = self.db_record.description
        else:
            val = await info.context.data_loaders.user_api_key_fields.load(
                (self.id, OrmApiKey.description),
            )
        return val

    @strawberry.field(description="The date and time the API key was created.")  # type: ignore
    async def created_at(
        self,
        info: Info[Context, None],
    ) -> datetime:
        if self.db_record:
            val = self.db_record.created_at
        else:
            val = await info.context.data_loaders.user_api_key_fields.load(
                (self.id, OrmApiKey.created_at),
            )
        return val

    @strawberry.field(description="The date and time the API key will expire.")  # type: ignore
    async def expires_at(
        self,
        info: Info[Context, None],
    ) -> Optional[datetime]:
        if self.db_record:
            val = self.db_record.expires_at
        else:
            val = await info.context.data_loaders.user_api_key_fields.load(
                (self.id, OrmApiKey.expires_at),
            )
        return val



from typing import Annotated, Optional, Union

import strawberry
from strawberry.relay import Node, NodeID
from typing_extensions import TypeAlias, assert_never

from phoenix.db import models
from phoenix.db.types.annotation_configs import (
    AnnotationType,
    OptimizationDirection,
)
from phoenix.db.types.annotation_configs import (
    CategoricalAnnotationConfig as CategoricalAnnotationConfigModel,
)
from phoenix.db.types.annotation_configs import (
    ContinuousAnnotationConfig as ContinuousAnnotationConfigModel,
)
from phoenix.db.types.annotation_configs import (
    FreeformAnnotationConfig as FreeformAnnotationConfigModel,
)


@strawberry.interface
class AnnotationConfigBase:
    name: str
    description: Optional[str]
    annotation_type: AnnotationType


@strawberry.type
class CategoricalAnnotationValue:
    label: str
    score: Optional[float]


@strawberry.type
class CategoricalAnnotationConfig(Node, AnnotationConfigBase):
    id_attr: NodeID[int]
    optimization_direction: OptimizationDirection
    values: list[CategoricalAnnotationValue]


@strawberry.type
class ContinuousAnnotationConfig(Node, AnnotationConfigBase):
    id_attr: NodeID[int]
    optimization_direction: OptimizationDirection
    lower_bound: Optional[float]
    upper_bound: Optional[float]


@strawberry.type
class FreeformAnnotationConfig(Node, AnnotationConfigBase):
    id_attr: NodeID[int]


AnnotationConfig: TypeAlias = Annotated[
    Union[CategoricalAnnotationConfig, ContinuousAnnotationConfig, FreeformAnnotationConfig],
    strawberry.union("AnnotationConfig"),
]


def _to_gql_categorical_annotation_config(
    annotation_config: models.AnnotationConfig,
) -> CategoricalAnnotationConfig:
    config = annotation_config.config
    assert isinstance(config, CategoricalAnnotationConfigModel)
    values = [
        CategoricalAnnotationValue(
            label=val.label,
            score=val.score,
        )
        for val in config.values
    ]
    return CategoricalAnnotationConfig(
        id_attr=annotation_config.id,
        name=annotation_config.name,
        annotation_type=config.type,
        optimization_direction=config.optimization_direction,
        description=config.description,
        values=values,
    )


def _to_gql_continuous_annotation_config(
    annotation_config: models.AnnotationConfig,
) -> ContinuousAnnotationConfig:
    config = annotation_config.config
    assert isinstance(config, ContinuousAnnotationConfigModel)
    return ContinuousAnnotationConfig(
        id_attr=annotation_config.id,
        name=annotation_config.name,
        annotation_type=config.type,
        optimization_direction=config.optimization_direction,
        description=config.description,
        lower_bound=config.lower_bound,
        upper_bound=config.upper_bound,
    )


def _to_gql_freeform_annotation_config(
    annotation_config: models.AnnotationConfig,
) -> FreeformAnnotationConfig:
    config = annotation_config.config
    assert isinstance(config, FreeformAnnotationConfigModel)
    return FreeformAnnotationConfig(
        id_attr=annotation_config.id,
        name=annotation_config.name,
        annotation_type=config.type,
        description=config.description,
    )


def to_gql_annotation_config(annotation_config: models.AnnotationConfig) -> AnnotationConfig:
    """
    Convert an SQLAlchemy AnnotationConfig instance to one of the GraphQL types.
    """
    config = annotation_config.config
    if isinstance(config, ContinuousAnnotationConfigModel):
        return _to_gql_continuous_annotation_config(annotation_config)
    if isinstance(config, CategoricalAnnotationConfigModel):
        return _to_gql_categorical_annotation_config(annotation_config)
    if isinstance(config, FreeformAnnotationConfigModel):
        return _to_gql_freeform_annotation_config(annotation_config)
    assert_never(annotation_config)



import json
from typing import Any, Optional, Union, cast

import strawberry
from strawberry import UNSET
from strawberry.scalars import JSON

from phoenix.db import models
from phoenix.db.types.model_provider import ModelProvider
from phoenix.server.api.helpers.prompts.models import (
    ContentPart,
    PromptChatTemplate,
    PromptMessage,
    PromptMessageRole,
    PromptTemplateFormat,
    PromptTemplateType,
    RoleConversion,
    TextContentPart,
    ToolCallContentPart,
    ToolCallFunction,
    ToolResultContentPart,
    normalize_response_format,
    normalize_tools,
    validate_invocation_parameters,
)


@strawberry.input
class ToolDefinitionInput:
    definition: JSON


@strawberry.input
class ResponseFormatInput:
    definition: JSON


@strawberry.input
class TextContentValueInput:
    text: str


@strawberry.input
class ToolResultContentValueInput:
    tool_call_id: str
    result: JSON


@strawberry.input
class ToolCallFunctionInput:
    type: Optional[str] = strawberry.field(default="function")
    name: str
    arguments: str


@strawberry.input
class ToolCallContentValueInput:
    tool_call_id: str
    tool_call: ToolCallFunctionInput


@strawberry.input(one_of=True)
class ContentPartInput:
    text: Optional[TextContentValueInput] = strawberry.UNSET
    tool_call: Optional[ToolCallContentValueInput] = strawberry.UNSET
    tool_result: Optional[ToolResultContentValueInput] = strawberry.UNSET


@strawberry.input
class PromptMessageInput:
    role: str
    content: list[ContentPartInput]


@strawberry.input
class PromptChatTemplateInput:
    messages: list[PromptMessageInput]


@strawberry.input
class ChatPromptVersionInput:
    description: Optional[str] = None
    template_format: PromptTemplateFormat
    template: PromptChatTemplateInput
    invocation_parameters: JSON = strawberry.field(default_factory=dict)
    tools: list[ToolDefinitionInput] = strawberry.field(default_factory=list)
    response_format: Optional[ResponseFormatInput] = None
    model_provider: ModelProvider
    model_name: str

    def __post_init__(self) -> None:
        self.invocation_parameters = {
            k: v for k, v in self.invocation_parameters.items() if v is not None
        }

    def to_orm_prompt_version(
        self,
        user_id: Optional[int],
    ) -> models.PromptVersion:
        tool_definitions = [tool.definition for tool in self.tools]
        tool_choice = cast(
            Optional[Union[str, dict[str, Any]]],
            cast(dict[str, Any], self.invocation_parameters).pop("tool_choice", None),
        )
        model_provider = ModelProvider(self.model_provider)
        tools = (
            normalize_tools(tool_definitions, model_provider, tool_choice)
            if tool_definitions
            else None
        )
        template = to_pydantic_prompt_chat_template_v1(self.template)
        response_format = (
            normalize_response_format(
                self.response_format.definition,
                model_provider,
            )
            if self.response_format
            else None
        )
        invocation_parameters = validate_invocation_parameters(
            self.invocation_parameters,
            model_provider,
        )
        return models.PromptVersion(
            description=self.description,
            user_id=user_id,
            template_type=PromptTemplateType.CHAT,
            template_format=self.template_format,
            template=template,
            invocation_parameters=invocation_parameters,
            tools=tools,
            response_format=response_format,
            model_provider=ModelProvider(self.model_provider),
            model_name=self.model_name,
        )


def to_pydantic_prompt_chat_template_v1(
    prompt_chat_template_input: PromptChatTemplateInput,
) -> PromptChatTemplate:
    return PromptChatTemplate(
        type="chat",
        messages=[
            to_pydantic_prompt_message(message) for message in prompt_chat_template_input.messages
        ],
    )


def to_pydantic_prompt_message(prompt_message_input: PromptMessageInput) -> PromptMessage:
    content = [
        to_pydantic_content_part(content_part) for content_part in prompt_message_input.content
    ]
    return PromptMessage(
        role=RoleConversion.from_gql(PromptMessageRole(prompt_message_input.role)),
        content=content,
    )


def to_pydantic_content_part(content_part_input: ContentPartInput) -> ContentPart:
    if content_part_input.text is not UNSET:
        assert content_part_input.text is not None
        return TextContentPart(
            type="text",
            text=content_part_input.text.text,
        )
    if content_part_input.tool_call is not UNSET:
        assert content_part_input.tool_call is not None
        return ToolCallContentPart(
            type="tool_call",
            tool_call_id=content_part_input.tool_call.tool_call_id,
            tool_call=ToolCallFunction(
                type="function",
                name=content_part_input.tool_call.tool_call.name,
                arguments=content_part_input.tool_call.tool_call.arguments,
            ),
        )
    if content_part_input.tool_result is not UNSET:
        assert content_part_input.tool_result is not None
        return ToolResultContentPart(
            type="tool_result",
            tool_call_id=content_part_input.tool_result.tool_call_id,
            tool_result=json.loads(cast(str, content_part_input.tool_result.result)),
        )
    raise ValueError("content part input has no content")
